/* analyzer.h -- declarations for dacopan-analyzer.  
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

#ifndef ANALYZER_H
#define ANALYZER_H

#include <arpa/inet.h>
#include <sys/time.h>

#include "config.h"


/* Protocols.  */
enum
{
  /*
   * DESIGN_CHANGES:
   * 2004-05-01 Andrew Salo
   *      Unknown protocol constant for each layer added.
   */

  PTU_IPV4,			/* Network layer protocol is IPv4.  */
  PTU_ARP,			/* Network layer protocol is ARP.  */
  PTU_NET_UNKNOWN,		/* Network layer protocol is unknown.  */
  PTU_TCP,			/* Transport layer protocol is TCP.  */
  PTU_UDP,			/* Transport layer protocol is UDP.  */
  PTU_TRANS_UNKNOWN,		/* Transport layer protocol is unknown.  */
  PTU_DNS,			/* Application layer protocol is DNS.  */
  PTU_HTTP,			/* Application layer protocol is HTTP.  */
  PTU_APP_UNKNOWN		/* Application layer protocol is unknown.  */
};

/* Number of protocols.  */
#define PTU_PROTOCOLS_NUM 9

/* Event types.  */
enum
{
  UNIT_SENT,
  UNIT_RECEIVED,
  UNIT_LOST
};

/* TCP states.  */
enum
{
  TCP_STATE_UNKNOWN
};

/* The name of the output file.  */
extern char *output_file_name;

/* The name of variables log file.  */
extern char *variables_log_name;

/* List of host structures.  */
extern struct host *host_list;

/* Number of packet trace files. */
extern int packet_trace_file_count;

/* List of link structures.  */
extern struct link *link_list;

/* List of flow structures.  */
extern struct flow *flow_list;

/* Application layer ports.  */
extern struct app_ports ports;

/* Exists ARP-layer flag.  */
extern short int arp_exist;

/* DESIGN_CHANGES:
 * 26-04-2004  Surikov Viktor
 *        Added variable for timediff;
 * 28-04-2004 Andrew Salo
 *       timediff changed to timealign.
 */
struct host
{
  struct in_addr ip;
  char *hostname;

  /* will be set by the PEF Writer */
  char *id;

  /* Packet trace file name.  */
  char *ptfname;

  /* Time alignment for this host.  */
  double timealign;

  /* the next element in a linked list */
  struct host *next;
};

/* ...  */
struct ptu
{
  /* DESIGN_CHANGES:
     2004-05-02 Mikhail Kryshen:
         trans_event and app_event fields are no longer needed.
  */

  /* Timestamp.  */
  struct timeval timestamp;

  /* Network layer.  */
  int net_type;
  void *net_hdr;

  /* Transport layer.  */
  int trans_type;
  void *trans_hdr;

  /* Application layer.  */
  int app_type;
  void *app_data;
  int app_len;

  /* Host.  */
  struct host *host;

  /* Links.  */
  struct ptu *next;

  /* Links received unit to corresponding sent unit.  */
  struct ptu *link;

  /* Not used now
     int sent_flag;
   */

  /* Corresponding events (used internally in eventcalc.c).  */
  struct event *net_event;
  /*  struct event *trans_event; */
  /*   struct event *app_event; */
};

struct link
{
  struct host *host1;
  struct host *host2;
  struct link *next;

  char *id;
  double mtu;
};

struct flow
{
  struct host *host1;
  int port1;

  struct host *host2;
  int port2;

  /* Will be set by the PEF Writer.  */
  char *id;

  /* The next element in a linked list.  */
  struct flow *next;
};

/* Protocol variables.  */
struct variables
  /* DESIGN_CHANGES:
     2004-05-02 Andrey Ananin:
         tt type is changed
	 from double to timeval.
  */
{
  /* Standard TCP state or TCP_STATE_UNKNOWN.  */
  int tcp_state;
  /* Congestion window (TCP).  */
  double cwnd;
  /* Slow start threshold (TCP).  */
  double ssthresh;
  /* Duplicated ack threshold (TCP).  */
  double dupack;
  /* Transition time (TCP).  */
  struct timeval tt;
  /* Retransmission timeout (TCP).  */
  double rto;
};

/* The events sequence is a linked list of elements of
   the structure defined below.  */
struct event
{
  /* DESIGN_CHANGES:
     2004-04-30 Mikhail Kryshen:
         added prev field;
	 added parent_index field.
     2004-05-01 Mikhail Kryshen
         renamed parent_index to tag.
     2004-05-02 Mikhail Kryshen
         added child field.
     2004-05-03 Mikhail Kryshen
         renamed tag to index.
     2004-05-05 Mikhail Kryshen
         added last_parent field for more effecient splitter algorithm.
  */

  /* Type of the event: UNIT_SENT, UNIT_RECEIVED, UNIT_LOST.  */
  int type;

  /* Unit id, will be set by the PEF Writer.  */
  char *id;

  struct host *host;
  struct link *link;
  struct flow *flow;

  /* Event time.  */
  struct timeval timestamp;

  /* Protocol type.  */
  int protocol;

  /* Protocol header - type of the referenced structure
     depends on the protocol field value.  */
  void *header;

  /* The unit data.  */
  void *data;

  /* Length of the unit data.  */
  int data_len;

  /* Protocol variables.  */
  struct variables vars;

  /* Child (higher-level) event.  */
  struct event *child;

  /* Parent (lower-level) event.  */
  struct event *parent;

  /* Last parent of the event (used internally in split.c).  */
  struct event *last_parent;

  /* Used in case of fragmentation to
     create a linked list of parents.  */
  struct event *next_parent;

  /* Corresponding sent event.  */
  struct event *unit_prev;

  /* Corresponding receive or lost event.  */
  struct event *unit_next;

  /* Source host (non-NULL for UNIT_RECEIVED) */
  struct host *host_prev;

  /* Destination host (non-NULL for UNIT_SENT) */
  struct host *host_next;

  /* Previous event in the sequence.  */
  struct event *prev;

  /* Next event in the sequence.  */
  struct event *next;

  /* Used internally in splitter to determine the first and the last
     sent or received fragment in the list of parents.  */
  unsigned int index;
};


/* This structure is used to store ports
   for application layer protocols.  */
struct app_ports
{
  u_int16_t http[MAX_PORTS];	/* List of HTTP protocol ports.  */
  int http_count;		/* HTTP protocol ports counter.  */

  u_int16_t dns[MAX_PORTS];	/* List of DNS protocol ports.  */
  int dns_count;		/* DNS protocol ports counter.  */
};

#endif /* ANALYZER_H */
