/* cmdlparse.c -- parse command line arguments for dacopan-analyzer.  
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Andrew Salo
   Date:   12.04.2004 */

#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "analyzer.h"
#include "config.h"
#include "cmdlparse.h"
#include "error.h"
#include "util.h"


/*
 * Print program version information.
 */
void
anlz_version (FILE *out)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-23 Andrew Salo
   *     The function is used instead of string.
   */

  fprintf (out, "%s %s\n", PACKAGE_NAME, VERSION);
}


/*
 * Print program help screen, if tip is set to zero.
 * Print short tip, otherwise.
 */
void
anlz_help (FILE *out, int tip)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-23 Andrew Salo
   *     The function is used instead of string.
   */

  extern char *program_invocation_name;	/* Program name.  */

  if (tip)
    {
      /* Show short tip.  */
      fprintf (out, "Try `%s --help' for more information.\n",
	       program_invocation_name);
      return;
    }

  /* Print help screen.  */
  /* Usage.  */
  fprintf (out, "Usage: %s [OPTIONS] FILE1 FILE2\n",
	   program_invocation_name);

  /* Options table.  */
  fprintf (out, "\nOptions:\n\
  -i, --ip-address=IP1[,IP2]     Specify IP addresses\n\
  -l, --var-log=FILE             Read variables from FILE (not supported)\n\
  -o, --output=FILE              Place the output into FILE\n\
  -t, --time-align=VAL1[,VAL2]   Set time alignments to hosts\n\
      --http=PORT1[,...]         Specify HTTP ports\n\
      --dns=PORT1[,...]          Specify DNS ports\n\
  -h, --help                     Show brief documentation and exit\n\
  -v, --version                  Show version information and exit\n");

  /* Bug report.  */
  fprintf (out, "\nReport bugs to <%s>.\n", PACKAGE_BUGREPORT);
}


/*
 * Parse IPv4 addresses in standard numbers-and-dots notation.
 *
 * str is IPv4 address representation.
 * ip_addrs is list of IP addresses.
 * ip_count is the pointer to IP address counter.
 */
void
parse_ip_addr (const char *str, struct in_addr ip_addrs[LOGS_MAX],
	       int *ip_count)
{
  const char delim[] = " ,";	/* IP addresses delimiters.  */
  char *buf;			/* Copy of str.  */
  char *ip;			/* IP address as a string.  */
  int count = *ip_count;
  int c;

  /* Copy the string containing IP addresses.  */
  buf = copy_data ((void *) str, strlen (str) + 1);

  ip = strtok (buf, delim);

  /* Parse IP addresses list.  */
  while (ip != NULL)
    {
      /* Check number of IP addresses.  */
      if (count > LOGS_MAX)
	{
	  /* Show error message and short tip.  */
	  warning (NULL, anlz_strerror (ERR_NUM_IP));
	  anlz_help (stderr, 1);
	  exit (EXIT_FAILURE);
	}

      /* Convert network address.  */
      c = inet_pton (AF_INET, ip, &ip_addrs[count]);

      if (c < 0)
	error (EXIT_FAILURE, NULL, strerror (errno));
      else if (c == 0)
	error (EXIT_FAILURE, ip, anlz_strerror (ERR_IP));

      /* Increase IP address counter.  */
      count++;

      /* Get next IP address.  */
      ip = strtok (NULL, delim);
    }

  /* New ip counter value.  */
  *ip_count = count;

  /* Free some memory.  */
  free (buf);
}


/*
 * Parse time alignments.
 *
 * str is time alignments representation.
 * time_align is list of time alignments.
 * time_count is the pointer to time alignments counter.
 */
void
parse_time_align (const char *str, double time_align[LOGS_MAX],
		  int *time_count)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-28 Andrew Salo
   *     Time alignments are represented as a list.
   */

  const char delim[] = " ,";	/* Time alignments delimiters.  */
  char *buf;			/* Copy of str.  */
  char *tstr;			/* Time alignment as a string.  */
  int count = *time_count;
  char *tmp;

  /* Copy the string containing IP addresses.  */
  buf = copy_data ((void *) str, strlen (str) + 1);

  tstr = strtok (buf, delim);

  /* Parse time alignments list.  */
  while (tstr != NULL)
    {
      /* Check number of time alignments.  */
      if (count > LOGS_MAX)
	{
	  /* Show error message and short tip.  */
	  warning (NULL, anlz_strerror (ERR_NUM_TIMEALIGN));
	  anlz_help (stderr, 1);
	  exit (EXIT_FAILURE);
	}

      errno = 0;

      time_align[count] = strtod (tstr, &tmp);

      /* Check the result.  */
      if (errno || tmp != (tstr + strlen (tstr)))
	error (EXIT_FAILURE, tstr, anlz_strerror (ERR_TIMEALIGN));

      /* Increase time alignments counter.  */
      count++;

      /* Get next time alignment.  */
      tstr = strtok (NULL, delim);
    }

  /* New ip counter value.  */
  *time_count = count;

  /* Free some memory.  */
  free (buf);
}


/*
 * Parse ports.
 *
 * str is ports representation.
 * ports is a vector containing ports.
 * ports_count is the pointer to ports counter.
 */
void
parse_app_ports (const char *str, u_int16_t ports[MAX_PORTS],
		 int *ports_count)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-30 Andrew Salo
   *     User may specify list of application layer ports in command line.
   */

  const char delim[] = " ,";	/* Ports delimiters.  */
  char *buf;			/* Copy of str.  */
  char *pstr;			/* Port as a string.  */
  int count = *ports_count;
  char *tmp;
  long int val;
  int i;

  /* Copy the string containing IP addresses.  */
  buf = copy_data ((void *) str, strlen (str) + 1);

  pstr = strtok (buf, delim);

  /* Parse ports list.  */
  while (pstr != NULL)
    {
      /* Check number of ports.  */
      if (count >= MAX_PORTS)
	{
	  /* Show error message and short tip.  */
	  warning (NULL, anlz_strerror (ERR_NUM_PORTS));
	  anlz_help (stderr, 1);
	  exit (EXIT_FAILURE);
	}

      errno = 0;

      val = strtol (pstr, &tmp, 10);

      /* Check the result.  */
      if (errno || tmp != (pstr + strlen (pstr)) || val < 0 || val > 65535)
	error (EXIT_FAILURE, pstr, anlz_strerror (ERR_PORT));

      ports[count] = (u_int16_t) val;

      /* Check if port is already in the list.  */
      for (i = 0, val = 0; i < count; i++)
	/* Set flag val if port is in the list.  */
	if (ports[i] == ports[count])
	  {
	    val = 1;
	    break;
	  }

      /* Increase ports counter if the port is unique.  */
      if (!val)
	count++;

      /* Get next port.  */
      pstr = strtok (NULL, delim);
    }

  /* New ports counter value.  */
  *ports_count = count;

  /* Free some memory.  */
  free (buf);
}


/*
 * Parse command line arguments of the Analyzer.
 */
void
parse_args (int argc, char **argv)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-28 Andrew Salo
   *     '--time-diff' changed to '--time-align'.
   * 2004-04-30 Andrew Salo
   *     User may specify list of application layer ports in command line.
   */

  /* Long options for getopt_long.  */
  static struct option long_options[] = {
    {"ip-address", required_argument, 0, 'i'},
    {"var-log", required_argument, 0, 'l'},
    {"output", required_argument, 0, 'o'},
    {"time-align", required_argument, 0, 't'},
    {"http", required_argument, 0, 0},
    {"dns", required_argument, 0, 0},
    {"help", no_argument, 0, 'h'},
    {"version", no_argument, 0, 'v'},
    {0, 0, 0, 0}
  };

  /* Short options for getopt_long.  */
  const char *short_options = "i:l:o:t:hv";

  int option_index;		/* Used by getopt_long.  */

  struct in_addr ip_addrs[LOGS_MAX];	/* List of IP addresses.  */
  int ip_count = 0;		/* IP address counter.  */

  double time_align[LOGS_MAX];	/* Vector of time alignments.  */
  int time_count = 0;		/* Time alignments counter.  */

  struct host *ptr = NULL;
  int i;

  /* Allow getopt printing error messages.  */
  opterr = 1;

  /* Parse options.  */
  while (1)
    {
      i = getopt_long (argc, argv, short_options, long_options,
		       &option_index);

      if (i == -1)
	/* No more options.  */
	break;

      switch (i)
	{
	case 0:
	  /* Long option has been specified.  */
	  /* Check for required argument --- at that moment
	     all options require and argument.  */
	  if (optarg == NULL)
	    error (EXIT_FAILURE, NULL, anlz_strerror (ERR_UNKNOWN));

	  /* Parse HTTP ports.  */
	  if (!strcmp ("http", long_options[option_index].name))
	    {
	      parse_app_ports (optarg, ports.http, &(ports.http_count));
	      break;
	    }

	  /* Parse DNS ports.  */
	  if (!strcmp ("dns", long_options[option_index].name))
	    {
	      parse_app_ports (optarg, ports.dns, &(ports.dns_count));
	      break;
	    }

	  /* Unknown error.  */
	  error (EXIT_FAILURE, NULL, anlz_strerror (ERR_UNKNOWN));
	  break;

	case 'i':
	  /* Parse IP address.  */
	  parse_ip_addr (optarg, ip_addrs, &ip_count);
	  break;

	case 'l':
	  /* Variables log.  */
	  variables_log_name = optarg;
	  error (EXIT_SUCCESS, NULL, 
		 anlz_strerror (WARN_VAR_LOGS_NOT_IMPL));
	  break;

	case 'o':
	  /* Output file name.  */
	  output_file_name = optarg;
	  break;

	case 't':
	  /* Parse time alignment.  */
	  parse_time_align (optarg, time_align, &time_count);
	  break;

	case 'h':
	  /* Print help page and successfully exit.  */
	  anlz_help (stdout, 0);
	  exit (EXIT_SUCCESS);
	  break;

	case 'v':
	  /* Print version info and successfully exit.  */
	  anlz_version (stdout);
	  exit (EXIT_SUCCESS);
	  break;

	case '?':
	  /* The error message is printed by getopt.  */
	  anlz_help (stderr, 1);
	  exit (EXIT_FAILURE);
	  break;

	default:
	  error (EXIT_FAILURE, NULL, anlz_strerror (ERR_UNKNOWN));
	}
    }

  /* Check number of IP addresses.  */
  if (ip_count < LOGS_MIN || ip_count > LOGS_MAX)
    {
      /* Show error message and short tip.  */
      warning (NULL, anlz_strerror (ERR_NUM_IP));
      anlz_help (stderr, 1);
      exit (EXIT_FAILURE);
    }

  /* Check number of time alignments.  */
  if (time_count > ip_count)
    {
      /* Show error message and short tip.  */
      warning (NULL, anlz_strerror (ERR_NUM_TIMEALIGN));
      anlz_help (stderr, 1);
      exit (EXIT_FAILURE);
    }

  /* Fill undefined time alignments with zeroes.  */
  for (; time_count < ip_count; time_count++)
    time_align[time_count] = 0;

  /* Check number of non-option arguments (packet trace files).  */
  packet_trace_file_count = argc - optind;

  if (packet_trace_file_count != ip_count)
    {
      /* Show error message and short tip.  */
      warning (NULL, anlz_strerror (ERR_NUM_FILES));
      anlz_help (stderr, 1);
      exit (EXIT_FAILURE);
    }

  /* Fill packet trace file names.  */
  for (i = 0, host_list = NULL; i < packet_trace_file_count; i++)
    {
      /* Create new host structure.  */
      if (host_list == NULL)
	{
	  /* Create first element in list.  */
	  ptr = (struct host *) malloc_check (sizeof (struct host));
	  host_list = ptr;
	}
      else
	{
	  /* Create non-first list element.  */
	  ptr->next = (struct host *) malloc_check (sizeof (struct host));
	  ptr = ptr->next;
	}

      /* Set IP address, time alignment, and packet trace file name.   */
      ptr->ip = ip_addrs[i];
      ptr->timealign = time_align[i];
      ptr->ptfname = argv[optind + i];

      /* Set other members to null.  */
      ptr->hostname = NULL;
      ptr->id = NULL;
      ptr->next = NULL;
    }
}
