/* ec_common.c -- common functions used in event calculator.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */


/* Author: Mikhail Kryshen
   Date:   04.05.2004 */

#include "analyzer.h"
#include "ec_common.h"
#include "ec_ip.h"
#include "ec_tcp.h"
#include "ec_udp.h"

struct protocol_module protocol_modules[PTU_PROTOCOLS_NUM];

struct protocol_module pm_ipv4;
struct protocol_module pm_tcp;
struct protocol_module pm_udp;

/*
 * Initialize protocol modules.  
 */
void
protocol_modules_init ()
{
  int i;
  for (i = 0; i < PTU_PROTOCOLS_NUM; i++)
    {
      protocol_modules[i].can_restore = 0;
      protocol_modules[i].restore_first = NULL;
      protocol_modules[i].restore_last = NULL;
      protocol_modules[i].unprocessed_first = NULL;
    }

  ipv4_module_init (&protocol_modules[PTU_IPV4]);
  tcp_module_init (&protocol_modules[PTU_TCP]);
  udp_module_init (&protocol_modules[PTU_UDP]);
}

/*
 * Set default variables.  
 */
void
set_default_vars (struct event *event)
{
  event->vars.tcp_state = 0;
  event->vars.cwnd = 0;
  event->vars.ssthresh = 0;
  event->vars.dupack = 0;
  event->vars.tt.tv_sec = 0;
  event->vars.tt.tv_usec = 0;
  event->vars.rto = 0;
}

/* 
 * Insert event to the linked list after the specified element.
 */
void
insert_event (struct event **first, struct event **last,
	      struct event *after, struct event *event)
{
  if (after == NULL)
    {
      /* Insert to the top of the list.  */
      event->prev = NULL;
      event->next = *first;
      if (first != NULL)
        *first = event;
    }
  else
    {
      event->prev = after;
      event->next = after->next;
      after->next = event;
    }

  if (event->next == NULL)
    {
      if (last != NULL)
        *last = event;
    }
  else
    {
      event->next->prev = event;
    }
}

/* 
 * Add event to the end of the linked list.
 */
void
add_event (struct event **first, struct event **last, struct event *event)
{
  insert_event (first, last, *last, event);
}

/*
 * Remove event from the linked list.  
 */
void
remove_event (struct event **first, struct event **last, struct event *event)
{
  if (event->prev == NULL)
    {
      *first = event->next;
    }
  else
    {
      event->prev->next = event->next;
      event->prev = NULL;
    }

  if (event->next == NULL)
    {
      *last = event->prev;
    }
  else
    {
      event->next->prev = event->prev;
      event->next = NULL;
    }
}

/* 
 * Add event to the end of the linked list of parents.  
 */
void
add_parent_event (struct event **first, struct event **last,
		  struct event *event)
{
  if (*last == NULL)
    *first = event;
  else
    (*last)->next_parent = event;

  event->next_parent = NULL;
  *last = event;
}

/*
 * Remove event from the linked list of parents.  
 */
void
remove_parent_event (struct event **first, struct event *prev,
		     struct event *event)
{
  if (prev == NULL)
    *first = event->next_parent;
  else
    prev->next_parent = event->next_parent;

  event->next_parent = NULL;
}
