/* ec_ip.c -- IP support for event splitter.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Mikhail Kryshen
   Date:   04.05.2004 */

#include <netinet/ip.h>
#include <netinet/in.h>

#include "analyzer.h"
#include "ec_common.h"

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/*
 * Check for the IP fragment of a transport-level event.  
 */
int
is_ip_fragment_of (struct event *event, struct event *fragment)
{
  /* Header of the first fragment.  */
  struct iphdr *first_header;
  /* Header of the fragment being processed.  */
  struct iphdr *current_header;

  if (fragment->protocol != PTU_IPV4 ||
      event->parent == NULL || event->parent->protocol != PTU_IPV4)
    {
      return 0;
    }

  first_header = (struct iphdr *) event->parent->header;
  current_header = (struct iphdr *) fragment->header;

  if (first_header->saddr == current_header->saddr &&
      first_header->daddr == current_header->daddr &&
      /* RFC791 says that protocols should match as well.  */
      first_header->protocol == current_header->protocol &&
      /* And, finally, check id's.  */
      first_header->id == current_header->id)
    {
      return 1;
    }

  return 0;
}

int
is_first_ip_fragment (struct event *event, struct ptu *source)
{
  return source->trans_hdr != NULL;
}

/*
 * Check if the specified fragment is the last fragment.
 */
int
is_last_ip_fragment (struct event *fragment)
{
  struct iphdr *ip = (struct iphdr *) fragment->header;

/* #ifdef DEBUG */
/*       fprintf (stderr, "*** %i | %i %i %i\n", */
/* 	      fragment->index, */
/* 	      (ip->frag_off & IP_MF), */
/* 	      IP_OFFMASK, IP_MF); */
/* #endif /\* DEBUG *\/ */

  return !(ip->frag_off & IP_MF);
}

/*
 * Check relation of 2 fragments.
 *
 * Returns FRAG_IMMEDIATE if fragment f2 immedeatly follows f1,
 * FRAG_FOLLOWS if fragment f2 follws f1 with some fragments between them,
 * FRAG_NONSEQ otherwise.  
 */
int
compare_ip_fragments (struct event *f1, struct event *f2)
{
  /* IP fragment headers.  */
  struct iphdr *ip1;
  struct iphdr *ip2;

  /* Offsets (in octets) extracted from IP headers.  */
  int ip1_offset;
  int ip2_offset;

  /* Flags extracted from IP headers.  */
  int ip1_flags;
  /* int ip2_flags; */

  if (f1 == NULL || f1->protocol != PTU_IPV4)
    return FRAG_NONSEQ;

  if (f2 == NULL)
    {
      /* Check if f1 is the last fragment.  */
      if (is_last_ip_fragment (f1))
	return FRAG_IMMEDIATE;
      else
	return FRAG_FOLLOWS;
    }

  if (f2->protocol != PTU_IPV4)
    return FRAG_NONSEQ;

  ip1 = (struct iphdr *) f1->header;
  ip1_flags = ip1->frag_off & ~IP_OFFMASK;
  ip1_offset = (ip1->frag_off & IP_OFFMASK) * 8;
  
  ip2 = (struct iphdr *) f2->header;
  ip2_offset = (ip2->frag_off & IP_OFFMASK) * 8;
  
/*   /\* DEBUG *\/ */
/*   fprintf (stderr, "*** %i %i | %i %i %i %i %i\n", */
/* 	  f1->index, */
/* 	  f2->index, */
/* 	  ip1_offset, ip1->tot_len, */
/* 	  ip1->ihl * 4, */
/* 	  ip1_flags & IP_MF, */
/* 	  ip2_offset); */
  
  /* f1 is the last fragment.  */
  if (!(ip1_flags & IP_MF))
    return FRAG_NONSEQ;
  
  if (ip1_offset + ip1->tot_len - (ip1->ihl * 4) == ip2_offset)
    return FRAG_IMMEDIATE;
  
  if (ip1_offset + ip1->tot_len - (ip1->ihl * 4) < ip2_offset)
    return FRAG_FOLLOWS;
  
  return FRAG_NONSEQ;
}

void 
ipv4_module_init (struct protocol_module *module)
{
  module->can_restore = 1;
  module->is_fragment_of = &is_ip_fragment_of;
  module->is_first_fragment = &is_first_ip_fragment;
  module->compare_fragments = &compare_ip_fragments;
}
