/* ec_tcp.c -- TCP support for event splitter.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Mikhail Kryshen
   Date:   04.05.2004 */

#include <netinet/tcp.h>
#include <netinet/in.h>

#include "analyzer.h"
#include "ec_common.h"

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/* Use only events of type UNIT_SENT for defragmentation.  */
/* Do not use this mode (bug 0022).  -- Mikhail  */
/* #define TCP_DEFRAG_BY_SENT */

/*
 * Get the previous event in the flow.  
 */
struct event *
flow_prev_event (struct event *current)
{
  struct event *event;

  for (event = current->prev; event != NULL; event = event->prev)
    {
      if (event->flow == current->flow)
	return event;
    }

  return NULL;
}

/*
 * Get the next event in the flow.  
 */
struct event *
flow_next_event (struct event *current)
{
  struct event *event;

  for (event = current->next; event != NULL; event = event->next)
    {
      if (event->flow == current->flow)
	return event;
    }

  return NULL;
}

/*
 * Returns 1 if the TCP event has positive data length or at least on of 
 * the syn, fin or rst flags set.
 */
int
is_bounding_event (struct event *event, struct event *start)
{
  struct tcphdr *tcp;

  if (event->protocol != PTU_TCP)
    return 0;
  
  if (event->type != start->type)
    return 0;

  tcp = (struct tcphdr *)event->header;

  if (event->data_len > 0)
    {
      return 1;
    }

  if (event->host_prev == start->host_prev &&
      event->host_next == start->host_next &&
      (tcp->syn || tcp->fin || tcp->rst))
    {
      return 1;
    }

  return 0;
}

struct event *
prev_bounding_event (struct event *event)
{
  struct event *current;

  for (current = flow_prev_event (event); current != NULL; 
       current = flow_prev_event (current))
    {
      if (is_bounding_event (current, event))
	return current;
    }
  
  return NULL;
}

struct event *
next_bounding_event (struct event *event)
{
  struct event *current;

  for (current = flow_next_event (event); current != NULL; 
       current = flow_next_event (current))
    {
      if (is_bounding_event (current, event))
	return current;
    }
  
  return NULL;
}

/*
 * Check for the TCP fragment of an application-level event.  
 */
int
is_tcp_fragment_of (struct event *event, struct event *fragment)
{
  /* Header of the first fragment.  */
  struct tcphdr *first_header;
  /* Header of the fragment being processed.  */
  struct tcphdr *current_header;

  if (fragment->protocol != PTU_TCP ||
      event->parent == NULL || event->parent->protocol != PTU_TCP)
    {
      return 0;
    }

#ifdef TCP_DEFRAG_BY_SENT
  if (event->type == UNIT_RECEIVED)
    {
/*       fprintf (stderr, "***** itfo %p %p\n", */
/* 	       (void *)fragment->unit_prev->child,  */
/* 	       (void *)event->parent->unit_prev->child); */

      if (fragment->unit_prev->child == event->parent->unit_prev->child)
	return 1;
      else
	return 0;
    }
#endif /* TCP_DEFRAG_BY_SENT */

  first_header = (struct tcphdr *) event->parent->header;
  current_header = (struct tcphdr *) fragment->header;

  if (event->parent->host_prev == fragment->host_prev &&
      event->parent->host_next == fragment->host_next &&
      first_header->source == current_header->source &&
      first_header->dest == current_header->dest &&
      fragment->data_len > 0)
    {
      return 1;
    }

  return 0;
}

int
is_first_tcp_fragment (struct event *event, struct ptu *source)
{
  struct tcphdr *tcp;
  struct event *bounding;

#ifdef DEBUG
  fprintf (stderr, "is_first_tcp_fragment: event=%p, event->type=%i, "
	   "event->unit_prev=%p\n", (void *)event, event->type,
	   (void*)event->unit_prev);
#endif /* DEBUG */

#ifdef TCP_DEFRAG_BY_SENT
  if (event->type == UNIT_RECEIVED)
    {
      event = event->unit_prev;
    }
#endif /* TCP_DEFRAG_BY_SENT */

  if (event->protocol != PTU_TCP)
    return 0;

  if (event->data_len == 0)
    return 0;

  tcp = (struct tcphdr *)event->header;

  if (tcp->syn)
    return 1;

  bounding = prev_bounding_event (event);
  
  if (bounding == NULL)
    return 0;

  tcp = (struct tcphdr *)bounding->header;

  if (bounding->data_len > 0)
    {
      if (bounding->host == event->host &&
	  bounding->host_prev == event->host_prev &&
	  bounding->host_next == event->host_next)
	{
	  return 0;
	}
      else
	{
	  return 1;
	}
    }
  else if (tcp->syn)
    {
      return 1;
    }
  
  return 0;
}

int
is_last_tcp_fragment (struct event *event)
{
  struct tcphdr *tcp;
  struct event *bounding;

#ifdef DEBUG
  fprintf (stderr, "is_last_tcp_fragment: event=%p\n", (void *)event);
#endif /* DEBUG */

#ifdef TCP_DEFRAG_BY_SENT
  if (event->type == UNIT_RECEIVED)
    {
      event = event->unit_prev;
    }
#endif /* TCP_DEFRAG_BY_SENT */

  if (event->protocol != PTU_TCP)
    return 0;

  if (event->data_len == 0)
    return 0;

  tcp = (struct tcphdr *)event->header;

  if (tcp->fin || tcp->rst)
    return 1;

  bounding = next_bounding_event (event);
  
  if (bounding == NULL)
    return 0;

  tcp = (struct tcphdr *)bounding->header;

  if (bounding->data_len > 0)
    {
      if (bounding->host == event->host &&
	  bounding->host_prev == event->host_prev &&
	  bounding->host_next == event->host_next)
	{
	  return 0;
	}
      else
	{
	  return 1;
	}
    }
  else if (tcp->fin || tcp->rst)
    {
      return 1;
    }
  
  return 0;
}

/*
 * Check relation of 2 fragments.
 *
 * Returns FRAG_IMMEDIATE if fragment f2 immedeatly follows f1,
 * FRAG_FOLLOWS if fragment f2 follws f1 with some fragments between them,
 * FRAG_NONSEQ otherwise.
 */
int
compare_tcp_fragments (struct event *f1, struct event *f2)
{
  /* TCP fragment headers.  */
  struct tcphdr *tcp1;
  struct tcphdr *tcp2;

  if (f1 == NULL || f1->protocol != PTU_TCP)
    return FRAG_NONSEQ;

  if (f2 == NULL)
    {
      /* Check if f1 is the last fragment.  */
      if (is_last_tcp_fragment (f1))
	return FRAG_IMMEDIATE;
      else
	return FRAG_FOLLOWS;
    }

#ifdef TCP_DEFRAG_BY_SENT
  if (f1->type == UNIT_RECEIVED)
    {
      f1 = f1->unit_prev;
      f2 = f2->unit_prev;
    }
#endif /* TCP_DEFRAG_BY_SENT */

  if (f2->protocol != PTU_TCP)
    return FRAG_NONSEQ;

  tcp1 = (struct tcphdr *) f1->header;
  tcp2 = (struct tcphdr *) f2->header;
    
  /*       /\* DEBUG *\/ */
  /*       fprintf (stderr, "*** %i %i | %u %u %u\n", f1->index, f2->index, */
  /* 	      tcp_len, tcp1->seq, tcp2->seq); */

  if (tcp1->seq + f1->data_len == tcp2->seq)
    return FRAG_IMMEDIATE;

  if (tcp2->seq - (tcp1->seq + f1->data_len) <
      (tcp1->seq + f1->data_len) - tcp2->seq)
    return FRAG_FOLLOWS;

  return FRAG_NONSEQ;
}

void
tcp_module_init (struct protocol_module *module)
{
  module->can_restore = 1;
  module->is_fragment_of = &is_tcp_fragment_of;
  module->is_first_fragment = &is_first_tcp_fragment;
  module->compare_fragments = &compare_tcp_fragments;
}
