/* error.c -- error processing for dacopan-analyzer.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Andrew Salo
   Date:   12.04.2004 */

#include <stdio.h>
#include <stdlib.h>

#include "error.h"


/* Descriptive error message strings list.  */
const char *err_list[] = {
  "success",                                                   /* 000 */
  "unknown error",                                             /* 001 */

  /* Command line parser errors.  */
  "invalid IP address",                                        /* 002 */
  "invalid number of IP addresses",                            /* 003 */
  "invalid number of packet trace files",                      /* 004 */
  "invalid number of time alignments",                         /* 005 */
  "invalid time alignment value",	                       /* 006 */
  "invalid number of ports",                                   /* 007 */
  "invalid port value",                                        /* 008 */

  /* Log reader errors.  */
  "unsupported data link type",                                /* 009 */
  "truncated packet",	                                       /* 010 */

  /* Message mapper errors.  */
  "packet sent time is greater that the received time",        /* 011 */

  /* Message mapper and event calculator errors.  */
  "could not determine source or destination address",         /* 012 */

  /* Event calculator warnings.  */
  "irrelevant unit (ignored)",                                 /* 013 */
  "events have the same timestamps (ignored)",                 /* 014 */
  "no sent event registered for received unit (ignored)",      /* 015 */
  
  /* Command line parser warnings.  */
  "host variables logs are not supported (ignored)"            /* 016 */
};


/* Number of strings in err_list[].  */
const int err_num = 17;


/*
 * Print error message and terminate the program if status is non-zero.
 */
void
error (int status, const char *cause, const char *descr)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-22 Andrew Salo
   *     If descr is NULL, "unknown error" string is displayed.
   */

  extern char *program_invocation_name;	/* Program name.  */

  /* Print error message.  */
  if (cause != NULL)
    fprintf (stderr, "%s: %s: %s\n", program_invocation_name, cause,
	     (descr == NULL) ? anlz_strerror (ERR_UNKNOWN) : descr);
  else
    fprintf (stderr, "%s: %s\n", program_invocation_name,
	     (descr == NULL) ? anlz_strerror (ERR_UNKNOWN) : descr);

  /* Exit if status is non-zero.  */
  if (status != EXIT_SUCCESS)
    exit (status);
}


/*
 * Print warning.
 */
void
warning (const char *cause, const char *descr)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-30 Andrew Salo
   *     Warning displaying function added.
   */

  error (EXIT_SUCCESS, cause, descr);
}


/*
 * Get descriptive error message string by its symbolic name.
 */
const char *
anlz_strerror (int err)
{
  /*
   * DESIGN CHANGES:
   * 2004-04-22 Andrew Salo
   *     If err is invalid, "unknown error" string is returned.
   */

  /* Return descriptive error message.  */
  if (err >= 0 && err < err_num)
    return err_list[err];
  else
    return err_list[ERR_UNKNOWN];
}
