/* eventcalc.c -- calculate protocol variables and events.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Ananin Andrew
   Date: 17.04.2004
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include "eventcalc.h"
#include "ec_common.h"
#include "split.h"
#include "analyzer.h"
#include "error.h"
#include "util.h"

/* Add dropped event in the events sequence.  */
void
add_dropped (struct event *event, struct event *last, struct timeval tmp)
{
  /* Dropped event.  */
  struct event *dropped = NULL;

  /* Create dropped event.  */
  dropped = malloc_check (sizeof (struct event));

  /* Include event in the events sequence with corresponding fields.  */
  dropped->unit_prev = event;
  event->unit_next = dropped;
  dropped->unit_next = NULL;
  dropped->type = UNIT_LOST;
  dropped->header = event->header;
  dropped->protocol = event->protocol;
  dropped->host_next = event->host_next;
  dropped->host_prev = event->host_prev;
  dropped->parent = NULL;
  dropped->child = NULL;
  dropped->next_parent = NULL;
  dropped->link = event->link;
  dropped->flow = event->flow;
  dropped->data = NULL;
  dropped->data_len = 0;

  /* Set variables.  */
  set_default_vars (dropped);

  /* Set timestamp.  */
  if (cmp_timevals (last->timestamp, tmp) == 1)
    {
      dropped->timestamp = last->timestamp;
    }
  else
    {
      dropped->timestamp.tv_sec = tmp.tv_sec;
      dropped->timestamp.tv_usec = tmp.tv_usec;
    }

/*   dropped->next = last->next;   */
/*   last->next = dropped; */
  insert_event (NULL, NULL, last, dropped);

}

/* Calculate and add dropped event.  */
void
calc_dropped (struct event *event, struct event *last_success,
	      struct event *next_success, struct event *last)
{
  /* Temporary calculating timestamp value.  */
  struct timeval tmp;

  /* Last element in the events sequence.  */
  struct event *last_es = NULL;

  /* Flight times.  */
  struct timeval flight_time1;
  struct timeval flight_time2;

  /* Current event.  */
  struct event *current;

/*   fprintf (stderr, "**** last_success=%p, next_success=%p ****\n", */
/*            (void *)last_success, (void *)next_success); */

  /* Calculate droptime.  */
  /* Nothing but dropped packets before.  */
  if (last_success == NULL)
    {

      /* Except for the last packet:
         lost(i) = send(i) + (send(i)-send(i-1))/2.  */
      if (event->next == NULL)
	{
	  if (last != NULL)
	    {
	      tmp = diff_timeval (event->timestamp, last->timestamp);
	      tmp.tv_sec = tmp.tv_sec / 2;
	      tmp.tv_usec = tmp.tv_usec / 2;
	      tmp = add_timeval (event->timestamp, tmp);
	      if (tmp.tv_sec == 0 && tmp.tv_usec == 0)
		{
		  tmp.tv_usec = 1;
		}
	    }
	  /* If there is only one packet in it was dropped.  */
	  else
	    {
	      tmp.tv_sec = DEFAULT_DROP_TIME_SEC;
	      tmp.tv_usec = DEFAULT_DROP_TIME_USEC;
	    }
	}
      /* In other case:  
         lost(i) = (send(i)+send(i+1))/2.  */
      else
	{
	  tmp = add_timeval (event->timestamp, event->next->timestamp);
	  tmp.tv_sec = tmp.tv_sec / 2;
	  tmp.tv_usec = tmp.tv_usec / 2;
	}
    }
  else
    {
      if (next_success == NULL)
	{
	  /* Calculate flight time.  */
	  flight_time1 =
	    diff_timeval (last_success->timestamp,
			  last_success->unit_next->timestamp);
	  tmp.tv_sec = flight_time1.tv_sec / 2;
	  tmp.tv_usec = flight_time1.tv_sec / 2;
	  tmp = add_timeval (event->timestamp, tmp);
	}
      else
	{
	  /* Take flight time of last successfully and
	     next successfully sended packets.  */
	  flight_time1 =
	    diff_timeval (last_success->unit_next->timestamp,
			  last_success->timestamp);
	  flight_time2 =
	    diff_timeval (next_success->unit_next->timestamp,
			  next_success->timestamp);

	  /* Take the average of these flight times.  
	     Then divide this number by 2.  */
	  tmp = add_timeval (flight_time1, flight_time2);
	  tmp.tv_sec = tmp.tv_sec / 4;
	  tmp.tv_usec = tmp.tv_usec / 4;
	  tmp = add_timeval (event->timestamp, tmp);
	  /* #ifdef DEBUG
	     fprintf(stderr, "drop event timestamp: sec=%lu, msec=%lu\n", tmp.tv_sec, tmp.tv_usec);
	     #endif */
	}
    }

#ifdef DEBUG
  fprintf (stderr, "drop event timestamp: sec=%lu, msec=%lu\n", tmp.tv_sec,
	  tmp.tv_usec);
  fprintf (stderr, "event timestamp: sec=%lu, msec=%lu\n", event->timestamp.tv_sec,
	  event->timestamp.tv_usec);
#endif

  last_es = event;

  /*  Add event in the event sequence by the timestamp. */
  for (current = event->next; current != NULL; current = current->next)
    {
      last_es = current;

      if (cmp_timevals (current->timestamp, tmp) == 1)
	{
	  /* Insert the dropped event into the events sequence.  */
	  add_dropped (event, current->prev, tmp);
	  return;
	}
    }
  /* Add dropped event at the end of events sequence.  */
  tmp.tv_sec = event->timestamp.tv_sec;
  tmp.tv_usec = event->timestamp.tv_usec;
  add_dropped (event, last_es, tmp);
}

/* Find next successfully sended packet.  */
struct event *
find_next_success (struct event *events_sequence)
{
  /* Current event. */
  struct event *event;

  /* Next successfully sended packet.  */
  struct event *next_success = NULL;

  /* Find the packet.  */
  for (event = events_sequence; event != NULL; event = event->next)
    {
      if (event->unit_next != NULL)
	{
	  if (event->type == UNIT_SENT
	      && event->unit_next->type == UNIT_RECEIVED)
	    {
	      /* Next successfully sended packet was found!  */
	      next_success = event;
	      break;
	    }
	}
    }
  return next_success;
}

/* Build dropped events for the events sequence.  */
void
build_dropped (struct event **events_sequence)
{
  /* Current event in the events sequence.  */
  struct event *event;

  /* Last successfully sended packet.  */
  struct event *last_success = NULL;

  /* Previous packet.  */
  struct event *last = NULL;

  /* Next sucessfully sended packet.  */
  struct event *next_success = NULL;

  /* Find dropped event.  */
  for (event = *events_sequence; event != NULL; event = event->next)
    {
      if (event->unit_next != NULL)
	{
	  if (event->type == UNIT_SENT
	      && event->unit_next->type == UNIT_RECEIVED)
	    {
	      /* Remember last successfully sended packet.  */
	      last_success = event;
	    }
	}
      else
	{
	  if (event->type == UNIT_SENT)
	    {

	      /* Find next successfully sended packet.  */
	      next_success = find_next_success (event);

	      /* Add dropped event.  */
	      calc_dropped (event, last_success, next_success, last);
	      /* fprintf(stderr, "********Bingo!******\n"); */
	    }
	  /* Remember last event.  */
	}
      last = event;
    }
}


/* Calculate necessary protocol variables. */
void
calculate (struct ptu *ptu_sequence,
	   struct event **events_sequence,
	   struct link **links, struct flow **flows, char const *add_log_var)
{

  /* Current event.  */
  struct event *event;

  /* Current link.  */
  struct link *link;

  /* Initialize protocols support.  */
  protocol_modules_init ();

  /* Split PTU sequence on the Events sequence.  */
  split (ptu_sequence, events_sequence, host_list, links, flows);

  if (events_sequence != NULL)
    {
      /* Build droped events.  */
      build_dropped (events_sequence);

      /* Calculate variables.  */
      for (event = *events_sequence; event != NULL; event = event->next)
	{

	  /* Calculate different types of variables (not implemented). */
	  /* In first version of the Analyzer there is a 
	     minimal set of calculations.  */
	  switch (event->type)
	    {
	    case UNIT_SENT:
	    case UNIT_RECEIVED:

	      /* Set default variables.  */
	      set_default_vars (event);

	      /* Calculate transition time if corresponding unit is exist.  */
	      if (event->unit_next != NULL)
		{
		  if (event->unit_next->type != UNIT_LOST)
		    {
		      event->vars.tt =
			diff_timeval (event->unit_next->timestamp,
				      event->timestamp);
		    }
		}
	      break;
	    default:
	      break;
	    }
	}

      /* Write MTU_VALUE.  */
      for (link = *links; link != NULL; link = link->next)
	{
	  link->mtu = MTU_VALUE;
	}
    }
}
