/* logread.c -- read pcap log files for dacopan-analyzer.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Andrew Salo
   Date:   15.04.2004 */

#include <errno.h>
#include <pcap.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/types.h>

#include "analyzer.h"
#include "util.h"
#include "error.h"
#include "logread.h"


/* Pointer to the beginning of packet trace presentation.  */
struct ptu *ptp_begin;

/* Pointer to the current packet trace unit.  */
struct ptu *ptp_tmp;

/*
 * DESIGN_CHANGES:
 * 2004-05-01 Andrew Salo
 *      Global variable containing data link type added.
 */

/* Used by callback function to determine data link type.  */
int data_link_type;


/*
 * Sets network layer protocol type to unknown.
 */
void
set_net_unknown (struct ptu *p)
{
  p->net_type = PTU_NET_UNKNOWN;
  p->net_hdr = NULL;
}


/*
 * Sets transport layer protocol type to unknown.
 */
void
set_trans_unknown (struct ptu *p)
{
  p->trans_type = PTU_TRANS_UNKNOWN;
  p->trans_hdr = NULL;
}


/*
 * Sets application layer protocol type to unknown.
 */
void
set_app_unknown (struct ptu *p)
{
  p->app_type = PTU_APP_UNKNOWN;
  p->app_data = NULL;
  p->app_len = 0;
}


/*
 * Read HTTP message.
 */
void
read_http (const u_char * packet, u_int len)
{
  /* Set application layer protocol type to HTTP.  */
  ptp_tmp->app_type = PTU_HTTP;

  /* Set application layer data length.  */
  ptp_tmp->app_len = len;

  /* Copy application layer related data.  */
  ptp_tmp->app_data = copy_data ((void *) packet, len);
}


/*
 * Read DNS message.
 */
void
read_dns (const u_char * packet, u_int len)
{
  /* Set application layer protocol type to DNS.  */
  ptp_tmp->app_type = PTU_DNS;

  /* Set application layer data length.  */
  ptp_tmp->app_len = len;

  /* Copy application layer related data.  */
  ptp_tmp->app_data = copy_data ((void *) packet, len);
}


/*
 * Read unknown application layer protocol data.
 */
void
read_app_unknown (const u_char * packet, u_int len)
{
  /* Unknown protocol --- keep the data.  */
  ptp_tmp->app_type = PTU_APP_UNKNOWN;

  /* Set application layer data length.  */
  ptp_tmp->app_len = len;

  /* Copy application layer related data.  */
  ptp_tmp->app_data = copy_data ((void *) packet, len);
}


/*
 * Read TCP segment header.
 */
void
read_tcp (const u_char * packet, u_int len)
{
  struct tcphdr *hdr;
  int hdrlen;
  int i;

  /* Set transport layer protocol type.  */
  ptp_tmp->trans_type = PTU_TCP;

  /* Get TCP header length. */
  hdr = (struct tcphdr *) packet;
  hdrlen = hdr->doff * sizeof (u_int32_t);

  /* Allocate TCP header memory.  */
  ptp_tmp->trans_hdr = copy_data ((void *) packet, hdrlen);

  /* Host byte order.  */
  hdr = (struct tcphdr *) ptp_tmp->trans_hdr;

  hdr->source = ntohs (hdr->source);
  hdr->dest = ntohs (hdr->dest);
  hdr->seq = ntohl (hdr->seq);
  hdr->ack_seq = ntohl (hdr->ack_seq);
  hdr->window = ntohs (hdr->window);
  hdr->check = ntohs (hdr->check);
  hdr->urg_ptr = ntohs (hdr->urg_ptr);

  /* Define application layer protocol type using ports.  */
  for (i = 0; i < ports.http_count; i++)
    if (hdr->source == ports.http[i] || hdr->dest == ports.http[i])
      {
	/* HTTP message.  */
	read_http (packet + hdrlen, len - hdrlen);
	return;
      }

  /* Read unknown application layer protocol data.  */
  read_app_unknown (packet + hdrlen, len - hdrlen);
}


/*
 * Read UDP segment header.
 */
void
read_udp (const u_char * packet, u_int len)
{
  struct udphdr *hdr;
  int hdrlen;
  int i;

  /* Set transport layer protocol type.  */
  ptp_tmp->trans_type = PTU_UDP;

  /* Get UDP header length.  */
  hdrlen = sizeof (struct udphdr);

  /* Allocate UDP header memory.  */
  ptp_tmp->trans_hdr = copy_data ((void *) packet, hdrlen);

  /* Host byte order.  */
  hdr = (struct udphdr *) ptp_tmp->trans_hdr;

  hdr->source = ntohs (hdr->source);
  hdr->dest = ntohs (hdr->dest);
  hdr->len = ntohs (hdr->len);
  hdr->check = ntohs (hdr->check);

  /* Define application layer protocol type using ports.  */
  for (i = 0; i < ports.dns_count; i++)
    if (hdr->source == ports.dns[i] || hdr->dest == ports.dns[i])
      {
	/* DNS message.  */
	read_dns (packet + hdrlen, len - hdrlen);
	return;
      }

  /* Read unknown application layer protocol data.  */
  read_app_unknown (packet + hdrlen, len - hdrlen);
}


/*
 * Read IPv4 datagram header.
 */
void
read_ipv4 (const u_char * packet, u_int len)
{
  struct iphdr *hdr;
  int hdrlen;

  /* Set network layer protocol type.  */
  ptp_tmp->net_type = PTU_IPV4;

  /* Get internet header length. */
  hdr = (struct iphdr *) packet;
  hdrlen = hdr->ihl * sizeof (u_int32_t);

  /* Allocate IPv4 header memory.  */
  ptp_tmp->net_hdr = copy_data ((void *) packet, hdrlen);

  /* Host byte order.  */
  hdr = (struct iphdr *) ptp_tmp->net_hdr;

  hdr->tot_len = ntohs (hdr->tot_len);
  hdr->id = ntohs (hdr->id);
  hdr->frag_off = ntohs (hdr->frag_off);
  hdr->check = ntohs (hdr->check);

  /* Check for fragmentation.  */
  if (hdr->frag_off & IP_OFFMASK)
    {
      /* Packet is fragmented and this fragment is not first.  */
      /* Transport layer data is not presented in the fragment.  */
      set_trans_unknown (ptp_tmp);

      /* Read unknown application layer protocol data.  */
      read_app_unknown (packet + hdrlen, hdr->tot_len - hdrlen);
    }
  else
    {
      /* Packet is not fragmented or the fragment is first.  */
      switch (hdr->protocol)
	{
	case IPPROTO_TCP:
	  /* Transport layer protocol is TCP.  */
	  read_tcp (packet + hdrlen, hdr->tot_len - hdrlen);
	  break;

	case IPPROTO_UDP:
	  /* Transport layer protocol is UDP.  */
	  read_udp (packet + hdrlen, hdr->tot_len - hdrlen);
	  break;

	default:
	  /* Set transport and application layers
	     protocol types to unknown.  */
	  set_trans_unknown (ptp_tmp);
	  set_app_unknown (ptp_tmp);
	  break;
	}
    }
}


/*
 * Read ARP header.
 */
void
read_arp (const u_char * packet, u_int len)
{
  struct arphdr *hdr;
  int hdrlen;

  /* Set network layer protocol type.  */
  ptp_tmp->net_type = PTU_ARP;

  /* Get ARP header length.  */
  hdr = (struct arphdr *) packet;
  hdrlen = sizeof (struct arphdr) + 2 * (hdr->ar_hln + hdr->ar_pln);

  /* Copy ARP header.  */
  ptp_tmp->net_hdr = copy_data ((void *) packet, hdrlen);

  /* Host byte order.  */
  hdr = (struct arphdr *) ptp_tmp->net_hdr;

  hdr->ar_hrd = ntohs (hdr->ar_hrd);
  hdr->ar_pro = ntohs (hdr->ar_pro);
  hdr->ar_op = ntohs (hdr->ar_op);

  /* Set transport and application layers protocol types to unknown.  */
  set_trans_unknown (ptp_tmp);
  set_app_unknown (ptp_tmp);
}


/*
 * Read ethernet header.
 */
void
read_ether (const u_char * packet, u_int len)
{
  struct ethhdr *hdr;
  int hdrlen;

  /* Get ethernet header length.  */
  hdrlen = ETH_HLEN;

  /* Get ethernet header.  */
  hdr = (struct ethhdr *) packet;

  /* Read network layer protocol information.  */
  switch (ntohs (hdr->h_proto))
    {
    case ETH_P_IP:
      /* IP datagram.  */
      read_ipv4 (packet + hdrlen, len - hdrlen);
      break;

    case ETH_P_ARP:
      /* ARP packet.  */
      read_arp (packet + hdrlen, len - hdrlen);
      break;

    default:
      /* Set all layers protocol types to unknown.  */
      set_net_unknown (ptp_tmp);
      set_trans_unknown (ptp_tmp);
      set_app_unknown (ptp_tmp);
      break;
    }
}


/*
 * Read linux sll header.
 */
void
read_linux_sll (const u_char * packet, u_int len)
{
  /*
   * DESIGN_CHANGES:
   * 2004-05-01 Andrew Salo
   *      Linux cooked sockets support added.
   */

  struct sll_header *hdr;
  int hdrlen;

  /* Get header length.  */
  hdrlen = SLL_HDR_LEN;

  /* Get header.  */
  hdr = (struct sll_header *) packet;

  /* Read network layer protocol information.  */
  switch (ntohs (hdr->sll_protocol))
    {
    case ETH_P_IP:
      /* IP datagram.  */
      read_ipv4 (packet + hdrlen, len - hdrlen);
      break;

    case ETH_P_ARP:
      /* ARP packet.  */
      read_arp (packet + hdrlen, len - hdrlen);
      break;

    default:
      /* Set all layers protocol types to unknown.  */
      set_net_unknown (ptp_tmp);
      set_trans_unknown (ptp_tmp);
      set_app_unknown (ptp_tmp);
      break;
    }
}


/*
 * Callback function for processing the packets.
 */
void
lr_callback (u_char * unused, const struct pcap_pkthdr *hdr,
	     const u_char * packet)
{
  /*
   * DESIGN_CHANGES:
   * 2004-05-01 Andrew Salo
   *      One callback function for all data link types.
   */

  struct ptu *last;

  /* Check captured packet length.  */
  if (hdr->caplen < hdr->len)
    error (EXIT_FAILURE, NULL, anlz_strerror (ERR_TRUNC_PKT));

  /* Save pointer to last element in ptp. */
  last = ptp_tmp;

  /* Create new ptu.  */
  ptp_tmp = (struct ptu *) malloc_check (sizeof (struct ptu));

  /* Set link pointer to NULL.  */
  ptp_tmp->link = NULL;

  /* Set ptu timestamp.  */
  ptp_tmp->timestamp = hdr->ts;

  /* Process the packet depending on data link type.  */
  switch (data_link_type)
    {
    case DLT_RAW:
      /* Process the packet beginning from IP header.  */
      read_ipv4 (packet, hdr->caplen);
      break;

    case DLT_EN10MB:
      /* Process the packet beginning from ethernet header.  */
      read_ether (packet, hdr->caplen);
      break;

    case DLT_LINUX_SLL:
      /* Process the packet beginning from linux sll header.  */
      read_linux_sll (packet, hdr->caplen);
      break;

    default:
      /* Not supported. */
      error (EXIT_FAILURE, NULL, anlz_strerror (ERR_DLT_NOTSUP));
      break;
    }

  /* Set next pointer to NULL.  */
  ptp_tmp->next = NULL;

/*   /\* Event calculator (splitter) requires the following fields */
/*      to be initially set to NULL.  -- Mikhail Kryshen *\/ */
/*   ptp_tmp->net_event = NULL; */
/*   ptp_tmp->trans_event = NULL; */
/*   ptp_tmp->app_event = NULL; */

  /* Add the ptu to ptp.  */
  if (ptp_begin == NULL)
    ptp_begin = ptp_tmp;
  else
    last->next = ptp_tmp;
}


/*
 * Set filter for the log reader.
 */
void
lr_setfilter (pcap_t * pcd)
{
  struct bpf_program bp;	/* Filter program.  */

#define FILTER_STRLEN (INET_ADDRSTRLEN*2 + 16)

  char str[FILTER_STRLEN];	/* Filter string.  */
  char ip1[INET_ADDRSTRLEN];	/* First IP address string.  */
  char ip2[INET_ADDRSTRLEN];	/* Second IP address string.  */
  struct in_addr addr;		/* IP address.  */

  /* Only 2 packet trace files are currently supported.  */
  if (packet_trace_file_count != 2)
    error (EXIT_FAILURE, "lr_setfilter", anlz_strerror (ERR_NUM_IP));

  /* Get and convert ip1.  */
  addr = host_list->ip;

  if (inet_ntop (AF_INET, &addr, ip1, INET_ADDRSTRLEN) == NULL)
    error (EXIT_FAILURE, "inet_pton", strerror (errno));

  /* Get and convert ip2.  */
  addr = (host_list->next)->ip;

  if (inet_ntop (AF_INET, &addr, ip2, INET_ADDRSTRLEN) == NULL)
    error (EXIT_FAILURE, "inet_pton", strerror (errno));

  /* Prepare filter string. */
  if (snprintf (str, FILTER_STRLEN, "src or dst %s or %s", ip1, ip2) == -1)
    error (EXIT_FAILURE, "snprintf", strerror (errno));

  /* Compile filter program.  */
  if (pcap_compile (pcd, &bp, str, 1, 0) == -1)
    error (EXIT_FAILURE, "pcap_compile", pcap_geterr (pcd));

  /* Set filter program.  */
  if (pcap_setfilter (pcd, &bp) == -1)
    error (EXIT_FAILURE, "pcap_setfilter", pcap_geterr (pcd));

  /* Free allocated memory.  */
  pcap_freecode (&bp);
}


/* 
 * Read packet trace file and store ptu.
 */
void
read_log (struct host *h, struct ptu **begin)
{
  pcap_t *pcd;			/* Packet capture descriptor.  */
  char errbuf[PCAP_ERRBUF_SIZE];	/* Pcap error text.  */

  /* Initialize ptp beginning pointer.  */
  ptp_begin = NULL;

  /* Open file for reading.  */
  pcd = pcap_open_offline (h->ptfname, errbuf);

  if (pcd == NULL)
    {
      /* Check if file name is already in errbuf.  */
      if (strncmp (errbuf, h->ptfname,
		   MIN (strlen (errbuf), strlen (h->ptfname))) == 0)
	error (EXIT_FAILURE, NULL, errbuf);
      else
	error (EXIT_FAILURE, h->ptfname, errbuf);
    }

  /* Set filter.  */
  lr_setfilter (pcd);

  /* Determine data link type.  */
  data_link_type = pcap_datalink (pcd);

  /* Process the packets depending on data link type.  */
  switch (data_link_type)
    {
    case DLT_RAW:
      /* Raw IP.  */
    case DLT_EN10MB:
      /* Ethernet.  */
    case DLT_LINUX_SLL:
      /* Linux cooked sockets.  */
      pcap_loop (pcd, -1, lr_callback, NULL);
      break;

    default:
      /* Not supported. */
      error (EXIT_FAILURE, NULL, anlz_strerror (ERR_DLT_NOTSUP));
      break;
    }

  /* Set link to host for each element on ptp.  */
  for (ptp_tmp = ptp_begin; ptp_tmp != NULL; ptp_tmp = ptp_tmp->next)
    ptp_tmp->host = h;

  /* Close the packet capture descriptor.  */
  pcap_close (pcd);

  *begin = ptp_begin;
}
