/* msgmap.c -- map packet trace units for dacopan-analyzer.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Victor Surikov & Kirill Kulakov  */

#include <stdlib.h>
#include <netinet/ip.h>
#include <net/if_arp.h>

#include "msgmap.h"
#include "util.h"
#include "error.h"
#include "analyzer.h"

#include <stdio.h>


/*
 * DESIGN CHANGES:
 * 2004-05-19 Andrew Salo
 *     Function mapping_messages() changed to map_messages().
 */

/* Auxiliary functions.  */
int cmp_ptu_info (struct ptu *ptu1, struct ptu *ptu2);
void fix_timestamps (struct ptu **ptps);
void fix_timestamps2 (struct ptu *ptp);
void set_links (struct ptu **ptps, int n);
struct ptu *merge_ptps (struct ptu **pme, int n);


/*
 * Main module function.
 */
struct ptu *
map_messages (struct ptu **ptps)
{
  struct ptu *ptp;

  /* Fixed timestamp on timealign in every PTU.  */
  fix_timestamps (ptps);
  /* Set links between identical PTUs in different PTPs.  */
  set_links (ptps, packet_trace_file_count);
  /* Return PTUS, after merging PTPs.  */
  ptp = merge_ptps (ptps, packet_trace_file_count);
  fix_timestamps2 (ptp);
  return ptp;
}

/* Internal module auxiliary functions.  */

/* Function compare net headers two PTUs.  */
int
cmp_ptu_info (struct ptu *ptu1, struct ptu *ptu2)
{
  struct iphdr *iph1, *iph2;
  struct arphdr *arph1, *arph2;

  switch (ptu1->net_type)
    {
    case PTU_IPV4:
      iph1 = (struct iphdr *) ptu1->net_hdr;
      iph2 = (struct iphdr *) ptu2->net_hdr;
      if (iph1->id != iph2->id)
	return 0;
      break;

    case PTU_ARP:
      arph1 = (struct arphdr *) ptu1->net_hdr;
      arph2 = (struct arphdr *) ptu2->net_hdr;
      if (arph1->ar_hrd != arph2->ar_hrd)
	return 0;
      if (arph1->ar_pro != arph2->ar_pro)
	return 0;
      if (arph1->ar_hln != arph2->ar_hln)
	return 0;
      if (arph1->ar_pln != arph2->ar_pln)
	return 0;
      if (arph1->ar_op != arph2->ar_op)
	return 0;
      break;

    default:
      return 0;
    }
  return 1;
}

/* Auxiliary functions.  */

/*
 * Fix timestamp on timealign in every PTU.
 */
void
fix_timestamps (struct ptu **ptps)
{
  struct ptu *cur;
  int i;
  long diffsec;
  long diffusec;

  for (i = 0; i < packet_trace_file_count; i++)
    {
      for (cur = ptps[i]; cur != NULL; cur = cur->next)
	{
	  if (cur->net_type == PTU_ARP)
	    arp_exist = 1;
#ifdef DEBUG
	  fprintf (stderr, "fix_timestamps() ptu time=%f, difference=%f ",
		   TVTOD (cur->timestamp), cur->host->timealign);
#endif
	  diffsec = cur->host->timealign;
	  diffusec = (cur->host->timealign - (double) diffsec) * 1000000;
#ifdef DEBUG
	  fprintf (stderr, "diffusec=%lu ", diffusec);
#endif
	  if (diffusec + cur->timestamp.tv_usec < 0)
	    {
	      cur->timestamp.tv_usec += 1000000 + diffusec;
	      cur->timestamp.tv_sec--;
	    }
	  else if (diffusec + cur->timestamp.tv_usec > 1000000)
	    {
	      cur->timestamp.tv_usec += diffusec;
	      cur->timestamp.tv_usec -= 1000000;
	      cur->timestamp.tv_sec++;
	    }
	  else
	    {
	      cur->timestamp.tv_usec += diffusec;
	    }
	  cur->timestamp.tv_sec += diffsec;
#ifdef DEBUG
	  fprintf (stderr, "result ptu time=%lu.%lu\n", cur->timestamp.tv_sec,
		   cur->timestamp.tv_usec);
#endif
	}
    }
}

/*
 * Last fixing timestamps.  First timestamp equal zero.
 */
void
fix_timestamps2 (struct ptu *ptp)
{
  struct ptu *first = ptp;
  if (ptp == NULL)
    {
      return;
    }
  ptp = ptp->next;

  while (ptp != NULL)
    {
      ptp->timestamp = diff_timeval (ptp->timestamp, first->timestamp);
#ifdef DEBUG
      fprintf (stderr, "fix_timestamps2() timestamp=%f\n",
	       TVTOD (ptp->timestamp));
#endif
      ptp = ptp->next;
    }
#ifdef DEBUG
  fprintf (stderr, "fix_timestamps2() timestamp=%f\n",
	   TVTOD (first->timestamp));
#endif
  first->timestamp = diff_timeval (first->timestamp, first->timestamp);
}

/*
 * Links PTU packages between PTP lists.
 */
void
set_links (struct ptu **ptps, int n)
{
  struct ptu *cur1, *cur2;
  int i, j;
  struct in_addr ip1_src, ip1_dst, ip2_src, ip2_dst;

#ifdef DEBUG
  char strip1[INET_ADDRSTRLEN], strip2[INET_ADDRSTRLEN];
  struct iphdr *iph1;
#endif /* DEBUG */

  /* For every PTP.  */
  for (i = 0; i < n; i++)
    {
      /* For every PTU from changed PTP.  */
      for (cur1 = ptps[i]; cur1 != NULL; cur1 = cur1->next)
	{
	  /* If this packet was used, then use next packet.  */
#ifdef DEBUG
	  fprintf (stderr, "set_links() ptus=%u, ptu=%p, link=%p\n", i,
		   (void *) cur1, (void *) cur1->link);
#endif /* DEBUG */
	  if (cur1->link != NULL)
	    {
	      continue;
	    }

	  /* Cheked: current PTU package created for sent package.  */
	  if (get_ip_addrs (cur1->net_hdr, cur1->net_type,
			    &ip1_src, &ip1_dst) == -1)
	    {
	      error (EXIT_FAILURE, NULL, anlz_strerror (ERR_UNKNOWN_ADDRESS));
	    }
#ifdef DEBUG
	  inet_ntop (AF_INET, &ip1_src, strip1, INET_ADDRSTRLEN);
	  inet_ntop (AF_INET, &ip1_dst, strip2, INET_ADDRSTRLEN);
	  fprintf (stderr, "set_links() source IP=%s, dest IP=%s, time=%f\n",
		   strip1, strip2, TVTOD (cur1->timestamp));
	  inet_ntop (AF_INET, &(cur1->host->ip.s_addr), strip1,
		     INET_ADDRSTRLEN);
	  fprintf (stderr, "sent host IP=%s", strip1);
#endif /* DEBUG */
	  if (ip1_src.s_addr != cur1->host->ip.s_addr)
	    {
#ifdef DEBUG
	      fprintf (stderr, " packet does not sent from this host\n");
#endif /* DEBUG */
	      continue;		/* Packet not sent from current end-host.  */
	    }

	  /* Seach received host.  */
	  cur2 = NULL;
	  for (j = 0; j < n; j++)
	    {
	      if (j == i)
		{
		  continue;
		}
#ifdef DEBUG
	      inet_ntop (AF_INET, &(ptps[j]->host->ip.s_addr), strip1,
			 INET_ADDRSTRLEN);
	      fprintf (stderr, " receive host IP=%s", strip1);
#endif /* DEBUG */
	      if (ptps[j]->host->ip.s_addr == ip1_dst.s_addr)
		{
		  cur2 = ptps[j];
		  break;
		}
	    }
#ifdef DEBUG
	  fprintf (stderr, " equals with packet receive IP\n");
#endif /* DEBUG */
	  for (; cur2 != NULL; cur2 = cur2->next)
	    {
	      /* If this packet was used, then use next packet.  */
	      if (cur2->link != NULL)
		continue;
	      /* Compare packets.  */
	      if (cur1->net_type != cur2->net_type)
		continue;
	      if (get_ip_addrs (cur2->net_hdr, cur2->net_type,
				&ip2_src, &ip2_dst) == -1)
		error (EXIT_FAILURE, NULL, anlz_strerror (ERR_UNKNOWN_ADDRESS));
	      if (ip1_src.s_addr != ip2_src.s_addr
		  || ip1_dst.s_addr != ip2_dst.s_addr)
		continue;
#ifdef DEBUG
	      fprintf (stderr, "receive time=%f\n", TVTOD (cur2->timestamp));
#endif
	      if (cmp_ptu_info (cur1, cur2))
		{
		  if (cmp_timevals (cur1->timestamp, cur2->timestamp) == 1)
		    {
		      fprintf (stderr, "Try to add %lu.%06lu to"
			       " the time alignment for host %s\n",
			       diff_timeval (cur1->timestamp,
					     cur2->timestamp).tv_sec,
			       diff_timeval (cur1->timestamp,
					     cur2->timestamp).tv_usec,
			       ip2str (cur2->host->ip.s_addr));
		      error (EXIT_FAILURE, NULL,
			     anlz_strerror (ERR_MM_TIMESTAMP));
		    }
		  cur2->link = cur1;
#ifdef DEBUG
		  switch (cur1->net_type)
		    {
		    case PTU_IPV4:
		      iph1 = (struct iphdr *) cur1->net_hdr;
		      fprintf (stderr, "packet IP ID=%u linked!\n", iph1->id);
		      break;

		    case PTU_ARP:
		      fprintf (stderr, "packet ARP linked!\n");
		      break;

		    default:
		      fprintf (stderr, "packet UNKNOWN linked!\n");
		    }
#endif
		  break;
		}
	    }
	}
    }
}

/*
 * The function merge several PTP structures to PTUS structure that
 * sorts in chronology order.
 */
struct ptu *
merge_ptps (struct ptu **pme, int n)
{
  struct ptu *first;		/* Pointer to the first element of
				   structure PTUS.  */
  struct ptu *cur;		/* Current element of general PTU-list
				   (PTU in future).  */
  struct ptu *tmp;		/* Tempary pointer to a PTU.  */
  struct ptu *last;

  int i, adding;

  /* Get first PTU from first list.  */
  first = NULL;
  for (i = 0; i < n; i++)
    {
      if (pme[i] != NULL)
	{
	  first = pme[i];
#ifdef DEBUG
	  fprintf (stderr, "merge_ptps() n=%u, first=%p\n", n, (void *) first);
#endif /* DEBUG */
	  pme[i] = pme[i]->next;
	  first->next = NULL;
	  break;
	}
    }
  if (first == NULL)
    return first;
  last = first;			/* set last element.  */

  /* view other PTU and make list.  */
  while (1)			/* view first elements in sequences.  */
    {
      adding = 0;
      for (i = 0; i < n; i++)
	{
	  if (pme[i] == NULL)
	    {
	      continue;
	    }
	  if (pme[i] != NULL && (pme[adding] == NULL
				 || cmp_timevals (pme[adding]->timestamp,
						  pme[i]->timestamp) == 1))
	    {
	      adding = i;
	    }
	}
      if (pme[adding] == NULL)
	break;
#ifdef DEBUG
      fprintf (stderr, "adding=%p\n", (void *) pme[adding]);
#endif /* DEBUG */
      if (cmp_timevals (first->timestamp, pme[adding]->timestamp) == 1)
	{
	  /* we find PTU less than first.  */
	  tmp = first;
	  first = pme[adding];	/* make current PTU as first.  */
	  pme[adding] = pme[adding]->next;
	  first->next = tmp;
#ifdef DEBUG
	  fprintf (stderr, "move first. first=%p\n", (void *) first);
#endif /* DEBUG */
	  continue;		/* check other PTU.  */
	}
      if (cmp_timevals (pme[adding]->timestamp, last->timestamp) == 1)
	{
	  /* we find PTU less than last.  */
	  tmp = last;
	  last = pme[adding];	/* make current PTU as last.  */
	  pme[adding] = pme[adding]->next;
	  tmp->next = last;
	  last->next = NULL;
#ifdef DEBUG
	  fprintf (stderr, "move last. last=%p\n", (void *) last);
#endif /* DEBUG */
	  continue;		/* check other PTU.  */
	}
      cur = first;		/* search place in all sequence.  */
#ifdef DEBUG
      fprintf (stderr, "current PTU=%p\n", (void *) cur);
#endif /* DEBUG */
      while (cur->next != NULL)	/* view other PTU in list and find needed place.  */
	{
/*#ifdef DEBUG
	  fprintf (stderr, "timestamps cur->next=%f adding=%f\n",
		   TVTOD (cur->next->timestamp),
		   TVTOD (pme[adding]->timestamp));
	  	  #endif*//* DEBUG */
	  if (cmp_timevals (cur->next->timestamp, pme[adding]->timestamp) ==
	      1)
	    {
#ifdef DEBUG
	      fprintf (stderr, "add PTU: first=%p added=%p last=%p\n",
		       (void *) cur, (void *) pme[adding],
		       (void *) cur->next);
#endif /* DEBUG */
	      tmp = cur->next;
	      cur->next = pme[adding];
	      pme[adding] = pme[adding]->next;
	      cur->next->next = tmp;
	      break;
	    }
	  cur = cur->next;
	}
      /* if it is a last PTU in list.  */
      if (cur->next == NULL)
	{
#ifdef DEBUG
	  fprintf (stderr, "add as last element\n");
#endif /* DEBUG */
	  cur->next = pme[adding];
	  pme[adding] = pme[adding]->next;
	  cur->next->next = NULL;
	  last = cur->next;
	}
    }

  return first;
}
