/* pefwrite.c -- write protocol events file for dacopan-analyzer.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Mikhail Kryshen & Viktor Surikov */

#include <errno.h>
#include <stdlib.h>		/* Used for EXIT_FAILURE.  */
#include <string.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <netinet/ip.h>
#include <net/if_arp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include "pefwrite.h"
#include "pefwrite_http.h"
#include "error.h"
#include "util.h"

#define INDENT_PREFIX "    "

#define HOST_ID_PREFIX "H"
#define LINK_ID_PREFIX "K"
#define FLOW_ID_PREFIX "F"
#define EVENT_ID_PREFIX "U"

char ip_buffer[16];

/*
 * Convert ip address to the human-readable format.
 */
char *
inaddr_format (struct in_addr ip)
{
  snprintf (ip_buffer, 16, "%i.%i.%i.%i",
	    (ip.s_addr >> 0) & 0xFF,
	    (ip.s_addr >> 8) & 0xFF,
	    (ip.s_addr >> 16) & 0xFF, 
	    (ip.s_addr >> 24) & 0xFF);
  return ip_buffer;
}

/*
 * Generate id from prefix and index.
 */
char *
create_id (char *prefix, int index)
{
  int id_len = strlen (prefix) + 11;
  char *id = malloc_check (id_len);
  snprintf (id, id_len, "%s%i", prefix, index);
  return id;
}

/*
 * Write the hosts section.  
 */
void
pef_write_hosts (FILE * out, struct host *hosts, char *prefix)
{
  /* ID counter.  */
  int id = 1;

  fprintf (out, "%s<hosts>\n", prefix);

  for (; hosts != NULL; hosts = hosts->next)
    {
      hosts->id = create_id (HOST_ID_PREFIX, id++);

      fprintf (out, "%s%s<host id=\"%s\" ip=\"%s\"",
	       prefix, INDENT_PREFIX, hosts->id, inaddr_format (hosts->ip));

      if (hosts->hostname != NULL)
	fprintf (out, "hostname=\"%s\"", hosts->hostname);

      fputs ("/>\n", out);
    }

  fprintf (out, "%s</hosts>\n", prefix);
}

/*
 * Write the links section.  
 */
void
pef_write_links (FILE * out, struct link *links, char *prefix)
{
  /* ID counter.  */
  int id = 1;

  fprintf (out, "%s<links>\n", prefix);

  for (; links != NULL; links = links->next)
    {
      links->id = create_id (LINK_ID_PREFIX, id++);

      fprintf (out, "%s%s<link id=\"%s\" host1=\"%s\" host2=\"%s\"/>\n",
	       prefix, INDENT_PREFIX, links->id,
	       links->host1->id, links->host2->id);
    }

  fprintf (out, "%s</links>\n", prefix);
}

/*
 * Write the flows section.  
 */
void
pef_write_flows (FILE * out, struct flow *flows, char *prefix)
{
  /* ID counter.  */
  int id = 1;

  fprintf (out, "%s<flows>\n", prefix);

  for (; flows != NULL; flows = flows->next)
    {
      flows->id = create_id (FLOW_ID_PREFIX, id++);

      fprintf (out, "%s%s<flow id=\"%s\" host1=\"%s\" port1=\"%i\" "
	       "host2=\"%s\" port2=\"%i\"/>\n",
	       prefix, INDENT_PREFIX, flows->id,
	       flows->host1->id, flows->port1,
	       flows->host2->id, flows->port2);
    }

  fprintf (out, "%s</flows>\n", prefix);
}

/*
 * Write the layers section.  
 */
void
pef_write_layers (FILE * out, char *prefix)
{
  fprintf (out, "%s<layers>\n", prefix);

  if (arp_exist == 1)
    {
      fprintf (out, "%s%s<layer id=\"L0\" name=\"arplayer\">\n", prefix,
	       INDENT_PREFIX);
      fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"arp\"/>\n",
	       prefix, INDENT_PREFIX, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s</layer>\n", prefix, INDENT_PREFIX);
    }

  fprintf (out, "%s%s<layer id=\"L1\" name=\"network\">\n", prefix,
	   INDENT_PREFIX);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"net_unknown\"/>\n",
	   prefix, INDENT_PREFIX, INDENT_PREFIX, PTU_NET_UNKNOWN);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"ipv4\"/>\n",
	   prefix, INDENT_PREFIX, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s</layer>\n", prefix, INDENT_PREFIX);

  fprintf (out, "%s%s<layer id=\"L2\" name=\"transport\">\n", prefix,
	   INDENT_PREFIX);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"trans_unknown\"/>\n",
	   prefix, INDENT_PREFIX, INDENT_PREFIX, PTU_TRANS_UNKNOWN);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"tcp\"/>\n", prefix,
	   INDENT_PREFIX, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"udp\"/>\n", prefix,
	   INDENT_PREFIX, INDENT_PREFIX, PTU_UDP);
  fprintf (out, "%s%s</layer>\n", prefix, INDENT_PREFIX);

  fprintf (out, "%s%s<layer id=\"L3\" name=\"application\">\n", prefix,
	   INDENT_PREFIX);
  fprintf (out,
	   "%s%s%s<protocol id=\"P%u\" name=\"application_unknown\"/>\n",
	   prefix, INDENT_PREFIX, INDENT_PREFIX, PTU_APP_UNKNOWN);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"http\"/>\n", prefix,
	   INDENT_PREFIX, INDENT_PREFIX, PTU_HTTP);
  fprintf (out, "%s%s%s<protocol id=\"P%u\" name=\"dns\"/>\n", prefix,
	   INDENT_PREFIX, INDENT_PREFIX, PTU_DNS);
  fprintf (out, "%s%s</layer>\n", prefix, INDENT_PREFIX);

  fprintf (out, "%s</layers>\n", prefix);
}

/*
 * Write the variables section.  
 */
void
pef_write_variables (FILE * out, char *prefix)
{
  int i;

  fprintf (out, "%s<variables>\n", prefix);

  /* Packets header information for IPv4.  */
  fprintf (out, "%s%s<variable name=\"tos\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s<variable name=\"tot_len\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s<variable name=\"id\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out,
	   "%s%s<variable name=\"flag_rf\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out,
	   "%s%s<variable name=\"flag_df\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out,
	   "%s%s<variable name=\"flag_mf\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out,
	   "%s%s<variable name=\"frag_off\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s<variable name=\"ttl\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s<variable name=\"protocol\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s<variable name=\"source_addr\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);
  fprintf (out, "%s%s<variable name=\"dest_addr\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_IPV4);

  /* Packets header information for ARP.  */
  if (arp_exist == 1)
    {
      fprintf (out, "%s%s<variable name=\"protocol\" protocol=\"P%u\" "
	       "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"opcode\" protocol=\"P%u\" "
	       "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"sender_ip\" protocol=\"P%u\" "
	       "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"target_ip\" protocol=\"P%u\" "
	       "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"sender_hrd\" protocol=\"P%u\" "
	       "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"target_hrd\" protocol=\"P%u\" "
	       "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"trans_time\" protocol=\"P%u\""
	       " scope=\"unit\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
      fprintf (out, "%s%s<variable name=\"sent_time\" protocol=\"P%u\""
               " scope=\"unit\"/>\n", prefix, INDENT_PREFIX, PTU_ARP);
    }

  /* Packets header information for TCP.  */
  fprintf (out, "%s%s<variable name=\"source_port\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"dest_port\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"seq\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"ack_seq\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"window\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"urg_pointer\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"flag_fin\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"flag_syn\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"flag_rst\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"flag_psh\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"flag_ack\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"flag_urg\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);
  fprintf (out, "%s%s<variable name=\"data_offset\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_TCP);

  /* Packets header information for UDP.  */
  fprintf (out, "%s%s<variable name=\"source_port\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_UDP);
  fprintf (out, "%s%s<variable name=\"dest_port\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_UDP);
  fprintf (out, "%s%s<variable name=\"len\" protocol=\"P%u\" "
	   "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX, PTU_UDP);

  /* Packets header information for HTTP.  */
  for (i = 0; http_variables[i] != NULL; i++)
    fprintf (out, "%s%s<variable name=\"%s\" protocol=\"P%u\" "
	     "scope=\"unit-field\"/>\n", prefix, INDENT_PREFIX,
	     http_variables[i], PTU_HTTP);

  /* Packets header information for DNS.  */
  /* TODO: implement (for the next version, not in 1.0).  */

  /* In this version output only transition time.  */
  fprintf (out,
	   "%s%s<variable name=\"trans_time\" protocol=\"P%u P%u P%u P%u "
	   "P%u P%u P%u\" scope=\"unit\"/>\n",
	   prefix, INDENT_PREFIX, PTU_NET_UNKNOWN, PTU_IPV4,
	   PTU_TRANS_UNKNOWN, PTU_TCP, PTU_UDP, PTU_APP_UNKNOWN, PTU_HTTP);

  /* Sent time.  */
  fprintf (out,
	   "%s%s<variable name=\"sent_time\" protocol=\"P%u P%u P%u P%u "
	   "P%u P%u P%u\" scope=\"unit\"/>\n",
	   prefix, INDENT_PREFIX, PTU_NET_UNKNOWN, PTU_IPV4,
	   PTU_TRANS_UNKNOWN, PTU_TCP, PTU_UDP, PTU_APP_UNKNOWN, PTU_HTTP);

  /* Unit data.  */
  fprintf (out,
	   "%s%s<variable name=\"data\" protocol=\"P%u P%u\" scope=\"unit\"/>\n",
	   prefix, INDENT_PREFIX, PTU_APP_UNKNOWN, PTU_HTTP);

  fprintf (out, "%s</variables>\n", prefix);
}

/*
 * Write string in file with adaptation.
 */
void
fprintf_data (FILE * out, char *str, unsigned long n)
{
  unsigned long i;
  for (i = 0; i < n; i++)
    {
      if (str[i] == ']' && (i + 2) < n
	  && str[i + 1] == ']' && str[i + 2] == '>')
	{
	  fprintf (out, "]]]><![CDATA[]>");
	  i += 2;
	}
      else
	{
	  if (isprint (str[i]))
	    {
	      if (str[i] == '\\')
		fputs ("\\\\", out);
	      else
		fputc (str[i], out);
	    }
	  else
	    fprintf (out, "\\0x%02x",
		     (unsigned short) (unsigned char) str[i]);
	}
    }
}

/*
 * Write the data in event section.
 */
void
pef_write_data (FILE * out, struct event *event)
{
/*   char *buffer = (char *)malloc (event->data_len); */
/*   defragment_data (event, buffer); */
/*   fprintf_data (out, buffer, event->data_len); */
/*   free (buffer); */

  if (event->parent == NULL)
    {
      /* Reached the lowest-level event.  */
      fprintf_data (out, event->data, event->data_len);
    }
  else
    {
      for (event = event->parent; event != NULL; event = event->next_parent)
	{
	  pef_write_data (out, event);
	}
    }
}

/*
 * Write the variables in event section.
 */
void
pef_write_event_vars (FILE * out, struct event *unit, char *prefix)
{
  struct iphdr *iph;
  struct arphdr *arph;
  struct tcphdr *tcph;
  struct udphdr *udph;

  struct in_addr snd, trg;	/* ARP sender and target IP addresses.  */
  char *sha, *tha;		/* ARP header sender and target hardware addresses.  */

  fprintf (out, "%s<value name=\"sent_time\">%lu.%06lu</value>\n", prefix,
	   unit->timestamp.tv_sec, unit->timestamp.tv_usec);
/*
  if (unit->unit_next != NULL && unit->unit_next->type != UNIT_LOST)
    fprintf (out, "%s<value name=\"received_time\">%lu.%06lu</value>\n",
	     prefix, unit->unit_next->timestamp.tv_sec,
	     unit->unit_next->timestamp.tv_usec);
  else
    fprintf (out, "%s<value name=\"received_time\">unknown</value>\n",
	     prefix);
*/
  if (unit->unit_next != NULL && unit->unit_next->type != UNIT_LOST)
    fprintf (out, "%s<value name=\"trans_time\">%lu.%06lu</value>\n", prefix,
	     unit->vars.tt.tv_sec, unit->vars.tt.tv_usec);
  else
    fprintf (out, "%s<value name=\"trans_time\">unknown</value>\n", prefix);

  if (unit->protocol == PTU_IPV4)
    {
      iph = (struct iphdr *) unit->header;
      fprintf (out, "%s<value name=\"tos\">%u</value>\n", prefix, iph->tos);
      fprintf (out, "%s<value name=\"tot_len\">%u</value>\n", prefix,
	       iph->tot_len);
      fprintf (out, "%s<value name=\"id\">%u</value>\n", prefix, iph->id);
      fprintf (out, "%s<value name=\"flag_rf\">%u</value>\n", prefix,
	       (iph->frag_off & IP_RF) ? 1 : 0);
      fprintf (out, "%s<value name=\"flag_df\">%u</value>\n", prefix,
	       (iph->frag_off & IP_DF) ? 1 : 0);
      fprintf (out, "%s<value name=\"flag_mf\">%u</value>\n", prefix,
	       (iph->frag_off & IP_MF) ? 1 : 0);
      fprintf (out, "%s<value name=\"frag_off\">%u</value>\n", prefix,
	       (iph->frag_off & IP_OFFMASK) * 8);
      fprintf (out, "%s<value name=\"ttl\">%u</value>\n", prefix, iph->ttl);
      fprintf (out, "%s<value name=\"protocol\">%u</value>\n", prefix,
	       iph->protocol);
      fprintf (out, "%s<value name=\"check\">%u</value>\n", prefix,
	       iph->check);

      fprintf (out, "%s<value name=\"source_addr\">%s</value>\n", prefix,
	       ip2str (iph->saddr));
      fprintf (out, "%s<value name=\"dest_addr\">%s</value>\n", prefix,
	       ip2str (iph->daddr));

    }
  if (unit->protocol == PTU_ARP)
    {
      arph = (struct arphdr *) unit->header;

      fprintf (out, "%s<value name=\"protocol\">%u</value>\n", prefix,
	       arph->ar_pro);
      fprintf (out, "%s<value name=\"opcode\">%u</value>\n", prefix,
	       arph->ar_op);

      /* Get sender and target IP addresses.  */
      get_ip_addrs (arph, PTU_ARP, &snd, &trg);

      fprintf (out, "%s<value name=\"sender_ip\">%s</value>\n", prefix,
	       inet_ntoa (snd));
      fprintf (out, "%s<value name=\"target_ip\">%s</value>\n", prefix,
	       inet_ntoa (trg));

      /* Get ARP hardware addresses.  */
      sha = (char *) malloc_check (3 * arph->ar_hln);
      tha = (char *) malloc_check (3 * arph->ar_hln);

      get_arp_hrd_addrs (arph, sha, tha, 3 * arph->ar_hln);

      fprintf (out, "%s<value name=\"sender_hrd\">%s</value>\n", prefix, sha);
      fprintf (out, "%s<value name=\"target_hrd\">%s</value>\n", prefix, tha);

      free (sha);
      free (tha);
    }
  if (unit->protocol == PTU_TCP)
    {
      tcph = (struct tcphdr *) unit->header;
      fprintf (out, "%s<value name=\"source_port\">%u</value>\n", prefix,
	       tcph->source);
      fprintf (out, "%s<value name=\"dest_port\">%u</value>\n", prefix,
	       tcph->dest);
      fprintf (out, "%s<value name=\"seq\">%u</value>\n", prefix, tcph->seq);
      fprintf (out, "%s<value name=\"ack_seq\">%u</value>\n", prefix,
	       tcph->ack_seq);
      fprintf (out, "%s<value name=\"window\">%u</value>\n", prefix,
	       tcph->window);
      fprintf (out, "%s<value name=\"urg_pointer\">%u</value>\n", prefix,
	       tcph->urg_ptr);
      fprintf (out, "%s<value name=\"flag_fin\">%u</value>\n", prefix,
	       tcph->fin);
      fprintf (out, "%s<value name=\"flag_syn\">%u</value>\n", prefix,
	       tcph->syn);
      fprintf (out, "%s<value name=\"flag_rst\">%u</value>\n", prefix,
	       tcph->rst);
      fprintf (out, "%s<value name=\"flag_psh\">%u</value>\n", prefix,
	       tcph->psh);
      fprintf (out, "%s<value name=\"flag_ack\">%u</value>\n", prefix,
	       tcph->ack);
      fprintf (out, "%s<value name=\"flag_urg\">%u</value>\n", prefix,
	       tcph->urg);
      fprintf (out, "%s<value name=\"data_offset\">%u</value>\n", prefix,
	       tcph->doff);
    }
  if (unit->protocol == PTU_UDP)
    {
      udph = (struct udphdr *) unit->header;
      fprintf (out, "%s<value name=\"source_port\">%u</value>\n", prefix,
	       udph->source);
      fprintf (out, "%s<value name=\"dest_port\">%u</value>\n", prefix,
	       udph->dest);
      fprintf (out, "%s<value name=\"len\">%u</value>\n", prefix, udph->len);
    }

  /* Write HTTP message header fields.  */
  if (unit->protocol == PTU_HTTP)
    pefwrite_http (out, unit, prefix);

  /* Write data  */
  if (unit->protocol == PTU_APP_UNKNOWN || unit->protocol == PTU_DNS)
    {
      fprintf (out, "%s<value name=\"data\"><![CDATA[", prefix);
      pef_write_data (out, unit);
      fprintf (out, "]]></value>\n");
    }
}

/*
 * Write the events section.
 */
void
pef_write_events (FILE * out, struct event *events, char *prefix)
{
  /* ID counter.  */
  int id = 1;

  struct event *cur_parent;

  char newpref[80] = "\0";
  strcat (newpref, prefix);
  strcat (newpref, INDENT_PREFIX);
  strcat (newpref, INDENT_PREFIX);

  fprintf (out, "%s<events>\n", prefix);

  for (; events != NULL; events = events->next)
    {
      if (events->type != UNIT_SENT && events->unit_prev != NULL)
	{
	  events->id = events->unit_prev->id;
	}

      if (events->id == NULL)
	events->id = create_id (EVENT_ID_PREFIX, id++);

/* #ifdef DEBUG */
/*       fprintf (out, "%s%s<!-- debug: %p: data_len=%i, unit_prev=%p, " */
/*                "unit_next=%p -->\n", prefix, INDENT_PREFIX, (void *)events,  */
/*                events->data_len, (void *)events->unit_prev,  */
/*                (void *)events->unit_next); */
/* #endif /\* DEBUG *\/ */

      switch (events->type)
	{
	case UNIT_SENT:
	  fprintf (out, "%s%s<unit_sent id=\"%s\" source=\"%s\" "
		   "destination=\"%s\" protocol=\"P%u\" time=\"%lu.%06lu\"",
		   prefix, INDENT_PREFIX, events->id, events->host->id,
		   events->host_next->id, events->protocol,
		   events->timestamp.tv_sec, events->timestamp.tv_usec);
	  if (events->parent != NULL)
	    {
	      if (events->parent->id == NULL)
		events->parent->id = create_id (EVENT_ID_PREFIX, id++);
	      fprintf (out, " children=\"%s", events->parent->id);
	      for (cur_parent = events->parent->next_parent;
		   cur_parent != NULL; cur_parent = cur_parent->next_parent)
		{
		  if (cur_parent->id == NULL)
		    cur_parent->id = create_id (EVENT_ID_PREFIX, id++);
		  fprintf (out, " %s", cur_parent->id);
		}
	      fputc ('\"', out);
	    }
	  if (events->flow != NULL)
	    fprintf (out, " flow=\"%s\"", events->flow->id);

	  fprintf (out, ">\n");
	  pef_write_event_vars (out, events, newpref);
	  fprintf (out, "%s%s</unit_sent>\n", prefix, INDENT_PREFIX);
	  break;
	case UNIT_RECEIVED:
	  fprintf (out,
		   "%s%s<unit_received id=\"%s\" time=\"%lu.%06lu\"/>\n",
		   prefix, INDENT_PREFIX, events->id,
		   events->timestamp.tv_sec, events->timestamp.tv_usec);
	  break;
	case UNIT_LOST:
	  fprintf (out, "%s%s<unit_dropped id=\"%s\" time=\"%lu.%06lu\"/>\n",
		   prefix, INDENT_PREFIX, events->id,
		   events->timestamp.tv_sec, events->timestamp.tv_usec);
	}
    }

  fprintf (out, "%s</events>\n", prefix);
}

/* 
 * Write the events sequence to the output stream in the PEF format.  
 */
void
pef_write (FILE * out, struct event *events, struct host *hosts,
	   struct link *links, struct flow *flows)
{
  time_t current_time;
  time (&current_time);

  /* Write the header.  */
  fputs ("<?xml version=\"1.0\"?>\n"
	 "<!DOCTYPE protocol_events SYSTEM "
	 "\"http://www.cs.helsinki.fi/group/dacopan/events.dtd\">\n\n", out);

  fprintf (out,
	   "<!--\n"
	   "%sCreated by: %s version %s.\n"
	   "%s%s"
	   "-->\n\n",
	   INDENT_PREFIX, PACKAGE_NAME, PACKAGE_VERSION,
	   INDENT_PREFIX, ctime (&current_time));

  fputs ("<protocol_events>\n\n", out);

  /* Write sections.  */
  pef_write_hosts (out, hosts, INDENT_PREFIX);
  fputc ('\n', out);
  pef_write_flows (out, flows, INDENT_PREFIX);
  fputc ('\n', out);
  pef_write_links (out, links, INDENT_PREFIX);
  fputc ('\n', out);
  pef_write_layers (out, INDENT_PREFIX);
  fputc ('\n', out);
  pef_write_variables (out, INDENT_PREFIX);
  fputc ('\n', out);
  pef_write_events (out, events, INDENT_PREFIX);

  fputs ("\n</protocol_events>\n", out);
}

/* 
 * Write the events sequence to the output file in the PEF format.  
 */
void
pef_writef (char *filename, struct event *events, struct host *hosts,
	    struct link *links, struct flow *flows)
{
  FILE *file = fopen (filename, "w");

  if (!file)
    error (EXIT_FAILURE, filename, strerror (errno));

  pef_write (file, events, hosts, links, flows);

  fclose (file);
}
