/* pefwrite_http.c -- parse HTTP headers for pefwrite.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Andrew Salo
   Date:   10.05.2004 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "analyzer.h"
#include "pefwrite.h"
#include "pefwrite_http.h"
#include "util.h"

#define CRLF "\r\n"


/*
 * These are symbolic names for HTTP message status line tokens.
 */
enum
{
  HTTP_METHOD = 0,
  HTTP_URL,
  HTTP_VERSION,
  HTTP_STATUS_CODE,
  HTTP_PHRASE,
  HTTP_HEADER_FIELDS
};


/*
 * Names for status line tokens and header fields.
 */
const char *http_variables[] = {
  "method",
  "url",
  "version",
  "status_code",
  "phrase",
  "connection",
  "content_length",
  "content_type",
  "cookie",
  "date",
  "host",
  "user_agent",
  NULL
};


/*
 * Write HTTP header fields.
 *
 * Return -1 on error, 0 otherwise.
 */
int
pefwrite_http_header_fields (FILE * out, char *options, const char *prefix,
			     int print)
{
  char *tmp;
  int i;

  /* Check if the next field starts from  CRLF (end of fields).  */
  while (strstr (options, CRLF) != options)
    {
      /* Try to find header field in the list.  */
      for (i = HTTP_HEADER_FIELDS; http_variables[i] != NULL; i++)
	if (strhttpstr (options, http_variables[i]) == options
	    && options[strlen (http_variables[i])] == ':')
	  {
	    /* Write header field.  */
	    if (print)
	      fprintf (out, "%s<value name=\"%s\">", prefix,
		       http_variables[i]);

	    /* Move to header field value.  */
	    options += strlen (http_variables[i]) + 1;

	    /* Remove leading spaces.  */
	    while (*options == ' ')
	      options++;

	    /* Write header field value.  */
	    while (*options != '\r')
	      {
		/* Check for unexpected whitespaces.  */
		if (strchr ("\n", *options) || *options == '\0')
		  return -1;

		if (print)
		  fputc (*options, out);

		options++;
	      }

	    if (print)
	      fprintf (out, "</value>\n");
	    break;
	  }

      /* Move to next header field.  */
      tmp = strstr (options, CRLF);

      if (tmp == NULL)
	return -1;
      else
	options = tmp + 2;
    }

  return 0;
}


/*
 * Write HTTP request message.
 *
 * Return -1 on error, 0 otherwise.
 */
int
pefwrite_http_request (FILE * out, char *message, const char *prefix,
		       int print)
{
  /* Extract method from status line.  */
  if (print)
    fprintf (out, "%s<value name=\"%s\">", prefix,
	     http_variables[HTTP_METHOD]);

  /* Get data up to ' ' character.  */
  while (*message != ' ')
    {
      /* Check for unexpected whitespaces.  */
      if (strchr ("\r\n", *message) || *message == '\0')
	return -1;

      if (print)
	fputc (*message, out);

      message++;
    }

  if (print)
    fprintf (out, "</value>\n");

  message++;

  /* Extract URI from status line.  */
  if (print)
    fprintf (out, "%s<value name=\"%s\">", prefix, http_variables[HTTP_URL]);

  /* Get data up to ' ' character.  */
  while (*message != ' ')
    {
      /* Check for unexpected whitespaces.  */
      if (strchr ("\r\n", *message) || *message == '\0')
	return -1;

      if (print)
	fputc (*message, out);

      message++;
    }

  if (print)
    fprintf (out, "</value>\n");

  message++;

  /* Extract version from status line.  */
  if (print)
    fprintf (out, "%s<value name=\"%s\">",
	     prefix, http_variables[HTTP_VERSION]);

  /* Get data up to '\r ' character.  */
  while (*message != '\r')
    {
      /* Check for unexpected whitespaces.  */
      if (strchr (" \n", *message) || *message == '\0')
	return -1;

      if (print)
	fputc (*message, out);

      message++;
    }

  if (print)
    fprintf (out, "</value>\n");

  message += 2;

  /* Write HTTP header fields.  */
  return pefwrite_http_header_fields (out, message, prefix, print);
}


/*
 * Write HTTP response message.
 *
 * Return -1 on error, 0 otherwise.
 */
int
pefwrite_http_response (FILE * out, char *message, const char *prefix,
			int print)
{
  /* Extract version from status line.  */
  if (print)
    fprintf (out, "%s<value name=\"%s\">",
	     prefix, http_variables[HTTP_VERSION]);

  /* Get data up to ' ' character.  */
  while (*message != ' ')
    {
      /* Check for unexpected whitespaces.  */
      if (strchr ("\r\n", *message) || *message == '\0')
	return -1;

      if (print)
	fputc (*message, out);

      message++;
    }

  message++;

  if (print)
    fprintf (out, "</value>\n");

  /* Extract response code from status line.  */
  if (print)
    fprintf (out, "%s<value name=\"%s\">", prefix,
	     http_variables[HTTP_STATUS_CODE]);

  /* Get data up to ' ' character.  */
  while (*message != ' ')
    {
      /* Check for unexpected whitespaces.  */
      if (strchr ("\r\n", *message) || *message == '\0')
	return -1;

      if (print)
	fputc (*message, out);

      message++;
    }

  message++;

  if (print)
    fprintf (out, "</value>\n");

  /* Extract string from status line.  */
  if (print)
    fprintf (out, "%s<value name=\"%s\">", prefix,
	     http_variables[HTTP_PHRASE]);

  /* Get data up to '\r' character.  */
  while (*message != '\r')
    {
      /* Check for unexpected whitespaces.  */
      if (strchr ("\n", *message) || *message == '\0')
	return -1;

      if (print)
	fputc (*message, out);

      message++;
    }

  message += 2;

  if (print)
    fprintf (out, "</value>\n");

  /* Write HTTP header fields.  */
  return pefwrite_http_header_fields (out, message, prefix, print);
}


/*
 * Print HTTP message body as raw data.
 *
 * Return -1 on error, 0 otherwise.
 */
int
pefwrite_http_body (FILE * out, char *message, int len, const char *prefix)
{
  char *body;
  int bodylen;

  /* Find beginning of HTTP message body.  */
  body = strstr (message, CRLF CRLF);

  if (body == NULL)
    return -1;

  body += 4;

  /* Calculate body length.  */
  bodylen = len - (body - message);

  /* Check HTTP message body length.  */
  if (bodylen < 0)
    return -1;

  /* Do not write data.  */
  if (bodylen == 0)
    return 0;

  /* Write data.  */
  fprintf (out, "%s<value name=\"data\"><![CDATA[", prefix);
  fprintf_data (out, body, bodylen);
  fprintf (out, "]]></value>\n");

  return 0;
}


/*
 * Write HTTP message header.
 */
void
pefwrite_http (FILE * out, struct event *unit, const char *prefix)
{
  char *buf;			/* Copy of message.  */

  /* Allocate memory to message.  */
  buf = (char *) malloc_check (unit->data_len + 1);
  defragment_data (unit, buf);
  buf[unit->data_len] = '\0';

  /* Check if status line begins from version.  */
  if (strhttpstr (buf, "HTTP/") == buf)
    {
      /* Write HTTP response message.  */
      if (pefwrite_http_response (out, buf, prefix, 0) == -1)
	{
	  /* Write raw data.  */
	  fprintf (out, "%s<value name=\"data\"><![CDATA[", prefix);
	  fprintf_data (out, buf, unit->data_len);
	  fprintf (out, "]]></value>\n");
	}
      else
	{
	  /* Write header and data.  */
	  pefwrite_http_response (out, buf, prefix, 1);

	  pefwrite_http_body (out, buf, unit->data_len, prefix);
	}
    }
  else
    {
      /* Write HTTP request message.  */
      if (pefwrite_http_request (out, buf, prefix, 0) == -1)
	{
	  /* Write raw data.  */
	  fprintf (out, "%s<value name=\"data\"><![CDATA[", prefix);
	  fprintf_data (out, buf, unit->data_len);
	  fprintf (out, "]]></value>\n");
	}
      else
	{
	  /* Write header and data.  */
	  pefwrite_http_request (out, buf, prefix, 1);

	  pefwrite_http_body (out, buf, unit->data_len, prefix);
	}
    }

  free (buf);
}
