/* split.c -- build events sequence, list of links and list flows
              for events calculator.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/* Author: Mikhail Kryshen, Andrey Ananin
   Date: 30.04.2004 */

/* DESIGN_CHANGES:
   2004-05-04 Mikhail Kryshen:
       Separated protocol-specific and general functions.
*/

#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/in.h>
#include <assert.h>

#include "analyzer.h"
#include "split.h"
#include "ec_common.h"
#include "util.h"
#include "error.h"

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

/* Events sequence.  */
struct event *es_first;
struct event *es_last;

struct host *hosts_first;
struct link *links_first;
struct flow *flows_first;

/* Last fragment index (chronologically).  */
unsigned int last_index;

void defragment_from (struct event *event, struct ptu *source);

/*
 * Add event to the beginning of the unprocessed list.
 */
void
add_unprocessed (struct event *event)
{
  struct protocol_module *pm;
  pm = &protocol_modules[event->protocol];

  event->next_parent = pm->unprocessed_first;
  pm->unprocessed_first = event;
}

void
remove_unprocessed (struct event *prev, struct event *event)
{
  struct protocol_module *pm;
  pm = &protocol_modules[event->protocol];

  remove_parent_event (&(pm->unprocessed_first), prev, event);
}

/*
 * Remove all events from unprocessed list.  
 */
void
clear_unprocessed ()
{
  struct protocol_module *pm;
  int i;
  
  for (i = 0; i < PTU_PROTOCOLS_NUM; i++)
    {
      pm = &protocol_modules[i];

      while (pm->unprocessed_first != NULL)
	{
	  remove_parent_event (&(pm->unprocessed_first), NULL,
			       pm->unprocessed_first);
	}
    }
}

void
add_restore (struct event *event)
{
  struct protocol_module *pm;
  pm = &protocol_modules[event->parent->protocol];

  add_event (&(pm->restore_first), &(pm->restore_last), event);
}

void
remove_restore (struct event *event)
{
  struct protocol_module *pm;
  pm = &protocol_modules[event->parent->protocol];

  remove_event (&(pm->restore_first), &(pm->restore_last), event);
}

/*
 * Remove all unrestored events (and their free memory).  
 */
void
clear_restore ()
{
  struct protocol_module *pm;
  struct event *event;
  int i;
  
  for (i = 0; i < PTU_PROTOCOLS_NUM; i++)
    {
      pm = &protocol_modules[i];

      while (pm->restore_first != NULL)
	{
	  event = pm->restore_first;
	  remove_event (&(pm->restore_first), &(pm->restore_last), event);
	  free (event);
	}
    }
}

int
is_first_fragment (struct event *fragment, struct ptu *source)
{
  if (!protocol_modules[fragment->protocol].can_restore)
    return 0;

  return protocol_modules[fragment->protocol]
    .is_first_fragment (fragment, source);
}

int
is_first_net_fragment (struct event *fragment, struct ptu *source)
{  
  return 
    fragment->protocol == PTU_IPV4 &&
    is_first_fragment (fragment, source);
}

int
is_first_trans_fragment (struct event *fragment, struct ptu *source)
{
  return 
    (fragment->protocol == PTU_UDP ||
     fragment->protocol == PTU_TCP) &&
    is_first_fragment (fragment, source);
}

int
is_fragment_of (struct event *event, struct event *fragment)
{
  if (event->type != fragment->type)
    return 0;

  if (!protocol_modules[fragment->protocol].can_restore)
    return 0;

  return protocol_modules[fragment->protocol]
    .is_fragment_of (event, fragment);
}

int
compare_fragments (struct event *f1, struct event *f2)
{
#ifdef DEBUG
  fprintf (stderr, "compare_fragments: f1=%p, f2=%p\n", 
	   (void *)f1, (void *)f2);
#endif /* DEBUG */

  if (!protocol_modules[f1->protocol].can_restore)
    return 0;

  return protocol_modules[f1->protocol]
    .compare_fragments (f1, f2);
}

/*
 * Checks if the list of parents is complete. 
 */
int
is_restored (struct event *event)
{
#ifdef DEBUG
  fprintf (stderr, "is_restored: event=%p, event->last_parent=%p\n", 
	   (void *)event, (void *)event->last_parent);
#endif /* DEBUG */

  return event->last_parent != NULL &&
    event->last_parent->next_parent == NULL &&
    compare_fragments (event->last_parent, NULL) == FRAG_IMMEDIATE;
}

/*
 * Insert fragment to the linked list of event parents.
 */
void
add_fragment (struct event *event, struct event *fragment)
{
  int relation = 0;
  int added = 0;
  int all_fragments = 1;
  struct event *parent;

#ifdef DEBUG
  fprintf (stderr, "add_fragment: event=%p, fragment=%p\n",
	   (void *)event, (void *)fragment);
#endif /* DEBUG */

  for (parent = event->last_parent; parent != NULL; 
       parent = parent->next_parent)
    {
      if (!added)
	{
	  relation = compare_fragments (parent, fragment);
	  
/* #ifdef DEBUG */
/* 	  fprintf (stderr, "add_fragment: for: parent=%p, fragment=%p, " */
/* 		   "relation=%i, all_fragments=%i\n",  */
/* 		   (void *)parent, (void *)fragment, relation, */
/* 		   all_fragments); */
/* #endif /\* DEBUG *\/ */

	  if (relation == FRAG_NONSEQ)
	    return;
	}

      if (compare_fragments (parent, parent->next_parent) == FRAG_IMMEDIATE)
	{
	  if (all_fragments && parent->next_parent != NULL)
	    {
	      event->last_parent = parent->next_parent;
	    }

	  if (!added && relation == FRAG_IMMEDIATE)
	    {
	      /* Fragment was already included (retransmission).  */
	      fragment->child = event;

	      /* TODO: add retransmitted parents to the list, but mark them
                 to avoid duplication of the data.  */

	      return;
	    }
	}
      else if (!added && (relation == FRAG_IMMEDIATE ||
			  (relation == FRAG_FOLLOWS &&
			   (parent->next_parent == NULL ||
			    compare_fragments (parent->next_parent,
					       fragment) == FRAG_NONSEQ))))
	{
	  /* Insert the fragment.  */
	  fragment->next_parent = parent->next_parent;
	  parent->next_parent = fragment;
	  fragment->child = event;
	  event->data_len += fragment->data_len;

	  if (all_fragments && relation == FRAG_IMMEDIATE)
	    event->last_parent = fragment;
	  else
	    all_fragments = 0;

	  added = 1;
	}
      else
	{
	  all_fragments = 0;
	}
    }
}

/* 
 * Add host to list of hosts.  
 */
struct host *
add_host (struct in_addr ip)
{
  /* Current host in hosts list.  */
  struct host *host;
  /* Last host in the list.  */
  struct host *last_host = NULL;
  /* Flag is checked if there is no need to include element in hosts list.  */
  int flag = 0;
  /* New host.  */
  struct host *tmp = NULL;
  /* Returned element.  */
  struct host *ret_host = NULL;

#ifdef DEBUG
  fprintf (stderr, "add_host\n");
#endif /* DEBUG */

  /* Check hosts list for the element with the same ip address.  */
  for (host = hosts_first; host != NULL; host = host->next)
    {
      /* Remember the last host. */
      last_host = host;
      if (host->ip.s_addr == ip.s_addr)
	{
	  flag = 1;
	  break;
	}
    }
  /* Include new element in the list.  */
  if (flag != 1)
    {
      tmp = (struct host *) malloc_check (sizeof (struct host));
      tmp->hostname = NULL;
      tmp->id = NULL;
      tmp->ip = ip;
      last_host->next = tmp;
      tmp->next = NULL;
      ret_host = tmp;
    }
  else
    {
      ret_host = last_host;
    }
  return ret_host;
}

/* Check new flow element in list of flows;
   return 0 if the flow element is new in flows list.  */
int
check_new_flow (struct event *events_sequence, struct flow *flows_first,
		int type, struct flow **last_flow)
{
  /* TODO: update to use protocol modules.  */

  /* Flag is checked if there is no need to include 
     new element in the flows list.  */
  int flag = 0;
  /* Current flow in the flows list.  */
  struct flow *flow;

  /* Include new element in the flows list for in case
     TCP or UDP protocols.  */
  switch (type)
    {
    case PTU_TCP:
      /* Try to find the same flow in flows list for TCP.  */
      for (flow = flows_first; flow != NULL; flow = flow->next)
	{
	  /* Remember last flow.  */
	  *last_flow = flow;
	  if (((events_sequence->host_next == flow->host1) &&
	       (((struct tcphdr *) events_sequence->header))->source ==
	       flow->port1) && ((events_sequence->host_prev == flow->host2)
				&&
				(((struct tcphdr *) events_sequence->
				  header))->dest == flow->port2))
	    {
	      flag = 1;
	      break;
	    }
	  if (((events_sequence->host_prev == flow->host1) &&
	       (((struct tcphdr *) events_sequence->header))->dest ==
	       flow->port1) && ((events_sequence->host_next == flow->host2)
				&&
				(((struct tcphdr *) events_sequence->
				  header))->source == flow->port2))
	    {
	      flag = 1;
	      break;
	    }
	}
      break;

    case PTU_UDP:
      /* Try to find the same flow in flows list for UDP.  */
      for (flow = flows_first; flow != NULL; flow = flow->next)
	{
	  /* Remember last flow.  */
	  *last_flow = flow;
	  if (((events_sequence->host_next == flow->host1) &&
	       (((struct udphdr *) events_sequence->header))->source ==
	       flow->port1) && ((events_sequence->host_prev == flow->host2)
				&&
				(((struct udphdr *) events_sequence->
				  header))->dest == flow->port2))
	    {
	      flag = 1;
	      break;
	    }
	  if (((events_sequence->host_prev == flow->host1) &&
	       (((struct udphdr *) events_sequence->header))->dest ==
	       flow->port1) && ((events_sequence->host_next == flow->host2)
				&&
				(((struct udphdr *) events_sequence->
				  header))->source == flow->port2))
	    {
	      flag = 1;
	      break;
	    }
	}
      break;

    default:
      flag = 0;
      break;
    }
  return flag;
}

void
set_link (struct event *event)
{
  /* Last element of the links list.  */
  struct link *links_last = NULL;
  /* Current link in the links list.  */
  struct link *link;
  /* New link.  */
  struct link *tmp;
  /* Flag is checked if there is no need
     to include new element in the links list.  */
  int flag;

#ifdef DEBUG
  fprintf (stderr, "set_link: event=%p\n", (void *)event);
#endif /* DEBUG */

  /* Link of events sequence element is NULL by default.  */
  event->link = NULL;
  /* Add element in list if new.  */ 
  flag = 0;
  for (link = links_first; link != NULL; link = link->next)
    {
      links_last = link;
      if (((event->host_next == link->host1) &&
	   (event->host_prev == link->host2)) ||
	  ((event->host_next == link->host2) &&
	   (event->host_prev) == link->host1))
	{
	  flag = 1;
	  break;
	}
    }
  /* We should include new link element in the list.  */
  if (flag == 0)
    {
      tmp = (struct link *) malloc_check (sizeof (struct link));

      /* Add new element in the list.  */
      tmp->host1 = event->host_next;
      tmp->host2 = event->host_prev;
      if (links_last == NULL) 
	{
	  links_first = tmp; 
	}
      else
	{
	  links_last->next = tmp;
	}
      tmp->next = NULL;
      event->link = tmp;
    }
  else 
    {
      event->link = links_last;
    }
}

void
set_flow (struct event *event)
{
  /* Last element in the flows list.  */
  struct flow *last_flow = NULL;
  struct flow *tmp;
 
  /* Flow link of the events sequence is NULL by default.  */
  event->flow = NULL;

  /* Create flows list for UDP or TCP protocols.  */
  switch (event->protocol)
    {
    case PTU_TCP:
      /* Check for new element.  */

      if (!check_new_flow
	  (event, flows_first, PTU_TCP, &last_flow))
	{
	  tmp = (struct flow *) malloc_check (sizeof (struct flow));

	  /* Add new element in the list.  */
	  tmp->port1 = ((struct tcphdr *) event->header)->source;
	  tmp->port2 = ((struct tcphdr *) event->header)->dest;
	  tmp->host1 = event->host_next;
	  tmp->host2 = event->host_prev;
	  if (last_flow == NULL) 
	    {
	      flows_first = tmp;
	    }
	  else
	    {
	      last_flow->next = tmp;
	    }
	  tmp->next = NULL;
	  event->flow = tmp;
	}
      else
	{
	  event->flow = last_flow;
	}
      break;

    case PTU_UDP:
      /* Check for new element.  */
      if (!check_new_flow
	  (event, flows_first, PTU_UDP, &last_flow))
	{
	  tmp = (struct flow *) malloc_check (sizeof (struct flow));

	  /* Add new element in the list.  */
	  tmp->port1 = ((struct udphdr *) event->header)->source;
	  tmp->port2 = ((struct udphdr *) event->header)->dest;
	  tmp->host1 = event->host_next;
	  tmp->host2 = event->host_prev;
	  if (last_flow == NULL) 
	    {
	      flows_first = tmp;
	    }
	  else
	    {
	      last_flow->next = tmp;
	    }
	  tmp->next = NULL;
	  event->flow = tmp;
	}
      else
	{
	  event->flow = last_flow;
	}
      break;

    default:
      break;
    }
}

/* 
 * Create new event corresponding to the 
 * lowest (network) layer of the specified PTU.  
 */
struct event *
create_net_event (struct ptu *ptu)
{
  /* The event element in process.  */
  struct event *event;
  /* Source address */
  struct in_addr source;
  /* Destination address */
  struct in_addr destination;

#ifdef DEBUG
  fprintf (stderr, "create_net_event: ptu=%p\n", (void *)ptu);
#endif /* DEBUG */

  event = (struct event *) malloc_check (sizeof (struct event));
  set_default_vars (event);

  event->id = NULL;
  event->timestamp = ptu->timestamp;
  event->host = ptu->host;
  event->protocol = ptu->net_type;
  event->header = ptu->net_hdr;
  event->data = ptu->app_data;
  event->data_len = ptu->app_len;

  event->unit_next = NULL;
  event->unit_prev = NULL;

  event->child = NULL;
  event->parent = NULL;
  event->last_parent = NULL;
  event->next_parent = NULL;

  event->flow = NULL;

  event->index = last_index++;

  if (get_ip_addrs (ptu->net_hdr, ptu->net_type, &source, &destination) != 0)
    {
      error (EXIT_FAILURE, ptu->host->ptfname,
	     anlz_strerror (ERR_UNKNOWN_ADDRESS));
    }

  /* Determine the event type.  */
  if (event->host->ip.s_addr == source.s_addr)
    {
      event->type = UNIT_SENT;
      event->host_prev = event->host;
      event->host_next = add_host (destination);
    }
  else if (event->host->ip.s_addr == destination.s_addr)
    {
/*       /\* Ignore packets that was not sent.  *\/ */
/*       if (ptu->link == NULL) */
/* 	{ */
/* 	  error (EXIT_SUCCESS, ptu->host->ptfname, */
/* 		 anlz_strerror (WARN_NOT_SENT)); */

/* 	  free (event); */
/* 	  return NULL; */
/* 	} */

      event->type = UNIT_RECEIVED;
      event->host_prev = add_host (source);
      event->host_next = event->host;
    }
  else
    {
      /* Drop event which is not related to it's host.  */
      free (event);
      /* Display warning.  */
      error (EXIT_SUCCESS, ptu->host->ptfname,
	     anlz_strerror (WARN_IRRELEVANT_UNIT));
      return NULL;
    }

  ptu->net_event = event;

  if (ptu->link != NULL)
    {
      event->unit_prev = ptu->link->net_event;
      if (event->unit_prev != NULL)
	{
	  event->unit_prev->unit_next = event;

	  /* For some reason these fields may not be equal.  */
	  /* TODO: find out why that happens and (may be) remove
	     the line below.  */

	  /* 2004-05-10: fixed in the logread.c.  */
/* 	  event->data_len = event->unit_prev->data_len; */
/* 	  event->data = event->unit_prev->data; */
	}
    }

  /* Ignore packets that was not sent.  */
  if (event->type == UNIT_RECEIVED && event->unit_prev == NULL)
    {
#ifdef DEBUG      
      if (ptu->link != NULL)
	{
	  fprintf (stderr, "*** bug in msgmap? ptu=%p, ptu->link=%p ***\n",
		   (void *)ptu, (void *)ptu->link);
	}
#endif /* DEBUG */

      error (EXIT_SUCCESS, ptu->host->ptfname,
	     anlz_strerror (WARN_NOT_SENT));
      
      free (event);
      return NULL;
    }

  set_link (event);

  return event;
}

/* 
 * Create new event corresponding to the 
 * transport layer of the specified PTU.  
 */
struct event *
create_trans_event (struct ptu *ptu, struct event *parent)
{
  /* The event element in process.  */
  struct event *event;

#ifdef DEBUG
  fprintf (stderr, "create_trans_event: ptu=%p, parent=%p\n", 
	   (void *)ptu, (void *)parent);
#endif /* DEBUG */

  if (ptu->trans_hdr == NULL)
    return NULL;

  event = (struct event *) malloc_check (sizeof (struct event));
  set_default_vars (event);

  event->id = NULL;
  event->timestamp = ptu->timestamp;
  event->host = ptu->host;
  event->protocol = ptu->trans_type;
  event->header = ptu->trans_hdr;
  event->data = NULL;
  event->data_len = parent->data_len;

  event->type = parent->type;
  event->host_prev = parent->host_prev;
  event->host_next = parent->host_next;
  
  event->link = parent->link;

  event->index = 0;

  event->unit_next = NULL;
  event->unit_prev = NULL;

  event->parent = parent;
  event->last_parent = parent;
  event->next_parent = NULL;
  event->child = NULL;
  parent->child = event;

  set_flow (event);

  return event;
}

/* 
 * Create new event corresponding to the 
 * application layer of the specified PTU.  
 */
struct event *
create_app_event (struct ptu *ptu, struct event *parent)
{
  /* The event element in process.  */
  struct event *event;

#ifdef DEBUG
  fprintf (stderr, "create_app_event: ptu=%p, parent=%p\n", 
	   (void *)ptu, (void *)parent);
#endif /* DEBUG */

  event = (struct event *) malloc_check (sizeof (struct event));
  set_default_vars (event);

  event->id = NULL;
  event->timestamp = ptu->timestamp;
  event->host = ptu->host;
  event->protocol = ptu->app_type;
  event->header = NULL;
  event->data = NULL;
  event->data_len = parent->data_len;

  event->type = parent->type;
  event->host_prev = parent->host_prev;
  event->host_next = parent->host_next;

  event->link = parent->link;
  event->flow = parent->flow;

  event->index = 0;

  event->unit_next = NULL;
  event->unit_prev = NULL;

  /* Set the first parent.  */  
  event->parent = parent;
  event->last_parent = parent;
  event->next_parent = NULL;
  event->child = NULL;
  parent->child = event;

  return event;
}

int
defragment (struct event *child, struct event *parent,
	    struct event *prev_fragment)
{
#ifdef DEBUG
  fprintf (stderr, "defragment: child=%p, parent=%p\n", 
	   (void *)child, (void *)parent);
#endif /* DEBUG */

  if (is_fragment_of (child, parent))
    {
      remove_unprocessed (prev_fragment, parent);
      add_fragment (child, parent);
      return 1;
    }

  return 0;
}

/* 
 * Insert the defragmented event into the ES.  
 */
void
insert_restored (struct event *restored,
		 struct ptu *source)
{
  struct event *first_parent = NULL;
  struct event *last_parent = NULL;
  struct event *parent;

#ifdef DEBUG
  fprintf (stderr, "insert_restored: restored=%p, restored->protocol=%i, "
	   "restored->parent=%p\n", (void *)restored, 
	   restored->protocol, (void *)restored->parent);
#endif /* DEBUG */

  if (restored->parent != NULL &&
      restored->parent->unit_prev != NULL)
    {
      restored->unit_prev = restored->parent->unit_prev->child;

      if (restored->unit_prev != NULL)
	{
	  if (restored->type == UNIT_RECEIVED &&
	      restored->unit_prev->index == 0)
	    {
	      /* Only received event was defragmented.  */
	      /* PENDING: print warning?  */

#ifdef DEBUG
	      fprintf (stderr, "*** restored %p (UNIT_RECEIVED): "
		       "sent event was not defragmented. ***\n",
		       (void *)restored);
#endif /* DEBUG */
	      
	      free (restored);
	      return;
	    }

	  restored->unit_prev->unit_next = restored;
	}
    }

  /* Find chronologically first and last parents.  */
  for (parent = restored->parent; parent != NULL; 
       parent = parent->next_parent)
    {
      if (first_parent == NULL ||
	  parent->index < first_parent->index)
	{
	  first_parent = parent;
	}

      if (last_parent == NULL ||
	  parent->index > last_parent->index)
	{
	  last_parent = parent;
	}
    }

  if (restored->type == UNIT_SENT)
    {
      restored->index = first_parent->index;

      restored->timestamp = first_parent->timestamp;
      insert_event (&es_first, &es_last,
		    first_parent->prev, restored);
    }
  else
    {
      restored->index = last_parent->index;

      restored->timestamp = last_parent->timestamp;
      insert_event (&es_first, &es_last,
		    last_parent, restored);
    }
  
  defragment_from (restored, source);
}

/*
 * Create child event from it's first fragment (parent).
 */
struct event *
child_from_first (struct event *event, struct ptu *source)
{
#ifdef DEBUG
  fprintf (stderr, "child_from_first: event=%p\n", (void *)event);
#endif /* DEBUG */

  /* Is the first network-layer fragment? */
  if (is_first_net_fragment (event, source))      
    {
      /* Create transport-layer event from the first fragment.  */
      return create_trans_event (source, event);
    }
  
  if (is_first_trans_fragment (event, source))
    {
      return create_app_event (source, event);
    }

  add_unprocessed (event);
  return NULL;
}

/*
 * Search for child's parents (fragments) in the list of 
 * unprocessed fragments.  
 */
int
restore_from_unprocessed (struct event *child)
{
  struct event *prev;
  struct event *current;
  struct event *next;

#ifdef DEBUG
  fprintf (stderr, "restore_from_unprocessed: child=%p\n", (void *)child);
#endif /* DEBUG */

  if (is_restored (child))
    {
      /* Was not fragmented.  */
      return 1;
    }

  /* Check if other fragments of the event are 
     alreadey in the sequence.  */
  prev = NULL;
  current = protocol_modules[child->parent->protocol].unprocessed_first;
  while (current != NULL)
    {
#ifdef DEBUG
      fprintf (stderr, "restore_from_unprocessed: while: current=%p\n", 
	       (void *)current);
#endif /* DEBUG */

      next = current->next_parent;

      if (!defragment (child, current, prev))
	{
	  /* Current was not removed from unprocessed.  */
	  prev = current;
	}
      else if (is_restored (child))
	{
	  /* Defragmented.  */
	  return 1;
	}

      current = next;
    }

  /* Add to the list of not yet unfragmented events.  */
  add_restore (child);

  return 0;
}

void
defragment_from (struct event *event, struct ptu *source)
{
  struct event *child;
  struct event *next;
  struct protocol_module *pm;

#ifdef DEBUG
  fprintf (stderr, "defragment_from: event=%p\n", (void *)event);
#endif /* DEBUG */

  pm = &protocol_modules[event->protocol];

  child = child_from_first (event, source);
  
  if (child != NULL)
    {
      if (restore_from_unprocessed (child))
	insert_restored (child, source);
    }
  else
    {
      /* Event was not the first fragment, search it's child in the
	 list of events that already have first fragments.  */
      for (child = pm->restore_first; child != NULL; child = child->next)
	{
	  if (defragment (child, event, NULL))
	    break;
	}
    }

  /* Check if we had defragmented anything.  */
/*   for (child = restore_first; child != NULL; child = child->next) */
/*     { */
/*       if (is_restored (child)) */
/* 	{ */
/* 	  remove_event (&restore_first, &restore_last, child); */
/* 	  insert_restored (child, source); */
/* 	  break; */
/* 	} */
/*     } */

  child = pm->restore_first;
  while (child != NULL)
    {
      next = child->next;

      if (is_restored (child))
	{
	  remove_restore (child);
	  insert_restored (child, source);
	}

      child = next;
    }
}

/*
 * Create events sequence.  
 */
void
split (struct ptu *ptu_sequence,
       struct event **events_sequence, struct host *hosts,
       struct link **links, struct flow **flows)
{
  /* The sequence element in porcess.  */
  struct ptu *ptu;
  /* The event element in process.  */
  struct event *event;

#ifdef DEBUG
  fprintf (stderr, "split\n");
#endif /* DEBUG */

  /* Start from empty events sequence.  */
  es_first = NULL;
  es_last = NULL;

  hosts_first = hosts;
  links_first = NULL;
  flows_first = NULL;

  last_index = 1;

  for (ptu = ptu_sequence; ptu != NULL; ptu = ptu->next)
    {
      event = create_net_event (ptu);

#ifdef DEBUG
  fprintf (stderr, "split: for: ptu=%p, event=%p\n", 
	   (void *)ptu, (void *)event);
#endif /* DEBUG */

      /* Event was not created for some reason.  
         Warning was already printed in create_net_event ().  */
      if (event == NULL)
	continue;

      add_event (&es_first, &es_last, event);
      
      defragment_from (event, ptu);
    }
  
  clear_unprocessed();
  clear_restore();

  *events_sequence = es_first;
  *links = links_first;
  *flows = flows_first;
}
