/* util.c -- common utility functions for dacopan-analyzer.
   Copyright (C) 2004 Petrozavodsk State University.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/if_ether.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>

#include "analyzer.h"
#include "error.h"
#include "util.h"

/*
 * Allocates memory checking for errors.
 */
void *
malloc_check (int len)
{
  void *dest;			/* Pointer to newly allocated memory.  */

  /* Allocate memory.  */
  dest = malloc (len);

  if (dest == NULL)
    error (EXIT_FAILURE, NULL, strerror (errno));

  return dest;
}


/*
 * Allocates memory and copies data checking for errors.
 */
void *
copy_data (void *src, int len)
{
  void *dest;			/* Pointer to newly allocated memory.  */

  /* Allocate memory.  */
  dest = malloc_check (len);

  /* Copy data.  */
  memcpy (dest, src, len);

  return dest;
}

/* 
 * Gets source and destination IP addresses from
 * network layer protocol header.
 *
 * hdr is a pointer to network layer header,
 * type is network layer protocol type,
 * src and dest are pointers to structures need to be filled with
 * source and destination IP addresses.
 *
 * 0 is returned on success, -1 otherwise.
 */
int
get_ip_addrs (void *hdr, int type, struct in_addr *src, struct in_addr *dst)
{
  struct iphdr *ip_hdr;		/* IP datagram header.  */
  struct arphdr *arp_hdr;	/* ARP header.  */
  u_char *ptr;

  /* Check header.  */
  if (hdr == NULL)
    return -1;

  /* Extract IP addresses depending on protocol type.  */
  switch (type)
    {
    case PTU_IPV4:
      /* Network protocol layer type is IPv4.  */
      ip_hdr = (struct iphdr *) hdr;

      /* Fill source and destination addresses.  */
      src->s_addr = ip_hdr->saddr;
      dst->s_addr = ip_hdr->daddr;
      break;

    case PTU_ARP:
      /* Network protocol layer type is ARP.  */
      arp_hdr = (struct arphdr *) hdr;

      /* Check for IP protocol.  */
      if (arp_hdr->ar_pro != ETH_P_IP
	  || arp_hdr->ar_pln != sizeof (in_addr_t))
	return -1;

      /* Point to source IP address and copy it.  */
      ptr = (u_char *) arp_hdr + sizeof (struct arphdr) + arp_hdr->ar_hln;
      memcpy (&(src->s_addr), ptr, sizeof (in_addr_t));

      /* Point to destination IP address and copy it.  */
      ptr += arp_hdr->ar_pln + arp_hdr->ar_hln;
      memcpy (&(dst->s_addr), ptr, sizeof (in_addr_t));
      break;

    default:
      /* Network protocol layer type is unknown.  */
      return -1;
      break;
    }

  return 0;
}


/* 
 * Extracts sender and target hardware addresses from ARP header.
 *
 * hdr is a pointer to ARP header,
 * src and dest are pointers to strings need to be filled with
 * sender and target hardware addresses,
 * cnt is src and dst size.
 *
 * 0 is returned on success, -1 otherwise.
 */
int
get_arp_hrd_addrs (struct arphdr *hdr, char *snd, char *trg, size_t cnt)
{
  char *ptr;
  int len;
  int i;

  /* Calculate hardware address string length.  */
  len = 3 * hdr->ar_hln;

  if (len > cnt)
    return -1;

  /* Write sender hardware address.  */
  ptr = (char *) (hdr + 1);

  snprintf (snd, 3, "%02x", (unsigned char) *ptr++);
  snd += 2;

  for (i = 1; i < hdr->ar_hln; i++)
    {
      snprintf (snd, 4, ":%02x", (unsigned char) *ptr++);
      snd += 3;
    }

  /* Write target hardware address.  */
  ptr = (char *) (hdr + 1) + hdr->ar_hln + hdr->ar_pln;

  snprintf (trg, 3, "%02x", (unsigned char) *ptr++);
  trg += 2;

  for (i = 1; i < hdr->ar_hln; i++)
    {
      snprintf (trg, 4, ":%02x", (unsigned char) *ptr++);
      trg += 3;
    }

  return 0;
}


/*
 * Compare time as structure timeval.
 * Return 0 if times equal; -1 if first time smaller; +1 if first time biger.
 */
int
cmp_timevals (struct timeval time1, struct timeval time2)
{
  if (time1.tv_sec == time2.tv_sec)
    {
      if (time1.tv_usec == time2.tv_usec)
	return 0;
      if (time1.tv_usec > time2.tv_usec)
	return 1;
      return -1;
    }
  if (time1.tv_sec > time2.tv_sec)
    return 1;
  return -1;
}

/* Addition time as structure timeval.  */
struct timeval
add_timeval (struct timeval time1, struct timeval time2)
{
  struct timeval res;
  res.tv_sec = 0;
  res.tv_usec = time1.tv_usec + time2.tv_usec;
  if (res.tv_usec >= 1000000)
    {
      res.tv_usec -= 1000000;
      res.tv_sec = 1;
    }
  res.tv_sec += time1.tv_sec;
  res.tv_sec += time2.tv_sec;
  return res;
}

/* Time difference (module).  */
struct timeval
diff_timeval (struct timeval time1, struct timeval time2)
{
  struct timeval res;
  long usec;
  if (cmp_timevals (time1, time2) == -1)
    {
      usec = time2.tv_usec - time1.tv_usec;
      if (usec < 0)
	{
	  usec += 1000000;
	  time2.tv_sec--;
	}
      res.tv_usec = usec;
      res.tv_sec = time2.tv_sec - time1.tv_sec;
    }
  else
    {
      usec = (signed long) time1.tv_usec - time2.tv_usec;
      if (usec < 0)
	{
	  usec += 1000000;
	  time1.tv_sec--;
	}
      res.tv_usec = usec;
      res.tv_sec = time1.tv_sec - time2.tv_sec;
    }
  return res;
}

/*
 * Convert ip address as long int to string format.
 */
char ip_buffer[16];

char *
ip2str (u_int32_t ip_i)
{
  snprintf (ip_buffer, 16, "%u.%u.%u.%u",
	    (ip_i >> 0) & 0xFF, (ip_i >> 8) & 0xFF,
	    (ip_i >> 16) & 0xFF, (ip_i >> 24) & 0xFF);
  return ip_buffer;
}

/*
 * Defragment the event data. Returns number of bytes added to the buffer.
 */
int
defragment_data (struct event *event, char *buffer)
{
  int index = 0;

  if (event->parent == NULL)
    {
      /* Reached the lowest-level event.  */
      memcpy (buffer, event->data, event->data_len);
      return event->data_len;
    }

  for (event = event->parent; event != NULL; event = event->next_parent)
    {
      index += defragment_data (event, &buffer[index]);
    }

  return index;
}


/*
 * This function is used while parsing HTTP message header fields.
 * It checks if one string "begins" from another. "Begins" mean that if
 * the first string i-th character is a letter then the second string
 * i-th character should be the same letter (uppercase or lowercase), and
 * if the first string i-th character is '-', the second string i-th
 * character should be '_'.
 *
 * haystack is the string, which should "begin" with needle.
 *
 * Return haystack if it "begins" from needle, NULL otherwise.
 */
const char *
strhttpstr (const char *haystack, const char *needle)
{
  int i;
  int count;

  /* Some of the strings is NULL or length of haystack is
     less than length of needle.  */
  if (haystack == NULL || needle == NULL
      || strlen (haystack) < strlen (needle))
    return NULL;

  count = strlen (needle);

  for (i = 0; i < count; i++)
    {
      /* Compare letters ignoring the case.  */
      if (isalpha (haystack[i]))
	{
	  if (tolower (haystack[i]) == tolower (needle[i]))
	    continue;
	  else
	    return NULL;
	}

      /* Compare  '-' and '_'.  */
      if (haystack[i] == '-')
	{
	  if (needle[i] == '_')
	    continue;
	  else
	    return NULL;
	}

      /* Compare other symbols.  */
      if (haystack[i] == needle[i])
	continue;
      else
	return NULL;
    }

  return haystack;
}
