#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/types.h>

#include "analyzer.h"
#include "cmdlparse.h"
#include "error.h"
#include "logread.h"
#include "msgmap.h"
#include "util.h"
#include "config.h"

void
print_debug_info2 (struct ptu *begin)
{
  struct arphdr *arp_hdr;
  struct iphdr *ip_hdr;
  struct tcphdr *tcp_hdr;
  struct udphdr *udp_hdr;

  struct in_addr ip1, ip2;
  char strip1[INET_ADDRSTRLEN], strip2[INET_ADDRSTRLEN];

  char *data;
  int i;

  while (begin != NULL)
    {
      fprintf (stderr, "\nPTU addr=%p ", (void *) begin);

      fprintf (stderr, "timestamp=%u.%u ",
	      begin->timestamp.tv_sec, begin->timestamp.tv_usec);
      fprintf (stderr, "link=%p ", (void *) begin->link);
      if (inet_ntop (AF_INET, &(begin->host->ip), strip2, INET_ADDRSTRLEN) ==
	  NULL)
	error (EXIT_FAILURE, "inet_ntop", strerror (errno));
      fprintf (stderr, "host=%s\n", strip2);


      /* Print network layer protocol information.  */
      switch (begin->net_type)
	{
	case PTU_ARP:
	  /* Network layer protocol is ARP.  */
	  arp_hdr = (struct arphdr *) begin->net_hdr;

	  get_ip_addrs ((void *) arp_hdr, PTU_ARP, &ip1, &ip2);

	  if (inet_ntop (AF_INET, &ip1, strip1, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  if (inet_ntop (AF_INET, &ip2, strip2, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  /* Print ARP header fields and IP addresses.  */
	  fprintf (stderr, "ARP [ar_hrd=%u, ar_pro=0x%x, ar_hln=%u, "
		   "ar_pln=%u, ar_op=%u, ar_sip=%s, ar_tip=%s]\n",
		   arp_hdr->ar_hrd, arp_hdr->ar_pro, arp_hdr->ar_hln,
		   arp_hdr->ar_pln, arp_hdr->ar_op, strip1, strip2);
	  break;

	case PTU_IPV4:
	  /* Network layer protocol is IPV4.  */
	  ip_hdr = (struct iphdr *) begin->net_hdr;

	  get_ip_addrs ((void *) ip_hdr, PTU_IPV4, &ip1, &ip2);

	  if (inet_ntop (AF_INET, &ip1, strip1, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  if (inet_ntop (AF_INET, &ip2, strip2, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  /* Print IP header fields.  */
	  fprintf (stderr, "IPV4 [id=%u, frag_off=0x%x, ttl=%u, "
		   "protocol=%u, saddr=%s, daddr=%s]\n",
		   ip_hdr->id, ip_hdr->frag_off,
		   ip_hdr->ttl, ip_hdr->protocol, strip1, strip2);
	  break;

	default:
	  fprintf (stderr, "unknown\n");
	}

      /* Print transport layer protocol information.  */
      switch (begin->trans_type)
	{
	case PTU_TCP:
	  /* Transport layer protocol is TCP.  */
	  tcp_hdr = (struct tcphdr *) begin->trans_hdr;

	  /* Print TCP header fields.  */
	  fprintf (stderr, "TCP [source=%u, dest=%u, seq=%u, "
		   "ack_seq=%u]\n",
		   tcp_hdr->source, tcp_hdr->dest, tcp_hdr->seq,
		   tcp_hdr->ack_seq);
	  break;

	case PTU_UDP:
	  /* Transport layer protocol is UDP.  */
	  udp_hdr = (struct udphdr *) begin->trans_hdr;

	  /* Print TCP header fields.  */
	  fprintf (stderr, "UDP [source=%u, dest=%u, "
		   "len=%u, check=0x%x]\n",
		   udp_hdr->source, udp_hdr->dest, udp_hdr->len,
		   udp_hdr->check);
	  break;

	default:
	  fprintf (stderr, "unknown\n");
	}
      /* Print application layer protocol information.  */
      fprintf (stderr, "App: ");

      switch (begin->app_type)
	{
	case PTU_HTTP:
	  /* Application layer protocol is HTTP.  */
	  fprintf (stderr, "HTTP [len=%u]\n", begin->app_len);

	  data = (char *) begin->app_data;

	  /* Print part of HTTP message.  */
	  for (i = 0; i < begin->app_len; i++)
	    fputc (data[i], stderr);
	  break;

	case PTU_DNS:
	  /* Application layer protocol is DNS.  */
	  fprintf (stderr, "DNS [len=%u]\n", begin->app_len);
	  break;

	default:
	  fprintf (stderr, "unknown [len=%u]\n", begin->app_len);
	}
      /* Move to next element.  */
      begin = begin->next;
    }



}

void
print_debug_info (struct event *begin, struct host *hosts,
		  struct flow *flow_list, struct link *link_list)
{
  char strip1[INET_ADDRSTRLEN], strip2[INET_ADDRSTRLEN];

  while (begin != NULL)
    {
      fprintf (stderr, "\nEvent [addr=%p, index=%i,", (void *) begin,
	       begin->index);
      switch (begin->type)
	{
	case UNIT_SENT:
	  fprintf (stderr,
		   " type=UNIT_SENT, id=%p,\n host=%p, link=%p, flow=%p,\n",
		   (void *) begin->id, (void *) begin->host,
		   (void *) begin->link, (void *) begin->flow);
	  break;
	case UNIT_RECEIVED:
	  fprintf (stderr,
		   " type=UNIT_RECEIVED, id=%p,\n host=%p, link=%p, flow=%p,\n",
		   (void *) begin->id, (void *) begin->host,
		   (void *) begin->link, (void *) begin->flow);
	  break;
	case UNIT_LOST:
	  fprintf (stderr,
		   " type=UNIT_LOST, id=%p,\n host=%p, link=%p, flow=%p,\n",
		   (void *) begin->id, (void *) begin->host,
		   (void *) begin->link, (void *) begin->flow);
	  break;
	default:
	  fprintf (stderr,
		   " type=UNKNOWN, id=%p,\n host=%p, link=%p, flow=%p,\n",
		   (void *) begin->id, (void *) begin->host,
		   (void *) begin->link, (void *) begin->flow);
	  break;
	}
	switch(begin->protocol)
	{
	case PTU_IPV4:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=IPV4, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_ARP:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=ARP, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_NET_UNKNOWN:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=NET_UNKNOWN, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_TCP:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=TCP, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_UDP:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=UDP, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_TRANS_UNKNOWN:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=TRANS_UNKNOWN, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_DNS:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=DNS, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_HTTP:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=HTTP, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	case PTU_APP_UNKNOWN:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=APP_UNKNOWN, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	default:
      fprintf (stderr, " timestamp=%lu.%lu, protocol=UNKNOWN, header=%p, data=%p, data_len=%u",
	       begin->timestamp.tv_sec, begin->timestamp.tv_usec, (void *) begin->header, (void *) begin->data, begin->data_len);
	       break;
	       }
      fprintf (stderr, "\n variables:");
      fprintf (stderr,
	       " tcp_state=%u, cwnd=%.3f, ssthresh=%.3f, dupack=%.3f, tt=%lu.%lu, rto=%.3f",
	       begin->vars.tcp_state, begin->vars.cwnd, begin->vars.ssthresh,
	       begin->vars.dupack, begin->timestamp.tv_sec, begin->timestamp.tv_usec, begin->vars.rto);
      fprintf (stderr,
	       "\n parent=%p, next_parent=%p, child=%p\n unit_prev=%p, unit_next=%p,",
	       (void *) begin->parent, (void *) begin->next_parent, (void *) begin->child,
	       (void *) begin->unit_prev, (void *) begin->unit_next);
      fprintf (stderr, "\n host_prev=%p, host_next=%p, prev=%p]\n",
	       (void *) begin->host_prev, (void *) begin->host_next, (void *) begin->prev);
      begin = begin->next;
    }
  fprintf (stderr, "=========== hosts ======================\n");
  while (hosts != NULL)
    {
      if (inet_ntop (AF_INET, &(hosts->ip), strip2, INET_ADDRSTRLEN) == NULL)
	error (EXIT_FAILURE, "inet_ntop", strerror (errno));
      fprintf (stderr, "Host [addr=%p, ip=%s]\n", (void *) hosts, strip2);
      hosts = hosts->next;
    }

  while (link_list != NULL)
    {
      fprintf (stderr, "=========== link ======================\n");
      if (inet_ntop
	  (AF_INET, &(link_list->host1->ip), strip2, INET_ADDRSTRLEN) == NULL)
	error (EXIT_FAILURE, "inet_ntop", strerror (errno));
      fprintf (stderr, "link addr=%p\n", (void *) link_list);
      fprintf (stderr, "Host1 [addr=%p, ip=%s]\n", (void *) link_list->host1,
	       strip2);

      if (inet_ntop
	  (AF_INET, &(link_list->host2->ip), strip2, INET_ADDRSTRLEN) == NULL)
	error (EXIT_FAILURE, "inet_ntop", strerror (errno));
      fprintf (stderr, "Host2 [addr=%p, ip=%s]\n", (void *) link_list->host2,
	       strip2);
      fprintf (stderr, "next=%p, MTU= %.3f\n", (void *) link_list->next,
	       link_list->mtu);
      link_list = link_list->next;
    }

  while (flow_list != NULL)
    {
      fprintf (stderr, "=========== flow ======================\n");
      if (inet_ntop
	  (AF_INET, &(flow_list->host1->ip), strip2, INET_ADDRSTRLEN) == NULL)
	error (EXIT_FAILURE, "inet_ntop", strerror (errno));
      fprintf (stderr, "flow addr=%p\n", (void *) flow_list);
      fprintf (stderr, "Host1 [addr=%p, ip=%s ", (void *) flow_list->host1,
	       strip2);
      fprintf (stderr, "port=%u]\n", flow_list->port1);

      if (inet_ntop
	  (AF_INET, &(flow_list->host2->ip), strip2, INET_ADDRSTRLEN) == NULL)
	error (EXIT_FAILURE, "inet_ntop", strerror (errno));
      fprintf (stderr, "Host2 [addr=%p, ip=%s, ", (void *) flow_list->host2,
	       strip2);
      fprintf (stderr, "port=%u]\n", flow_list->port2);
      fprintf (stderr, "next=%p\n", (void *) flow_list->next);
      flow_list = flow_list->next;
    }
}


int
__wrap_main (int argc, char **argv)
{
  struct link *link_list = NULL;
  struct flow *flow_list = NULL;
  struct ptu **ptu_lists;
  struct host *current_host;
  struct ptu *ptp;
  struct event *es;
  int i;

  /* Calls CMD line parser module.  */
  parse_args (argc, argv);

  /* Make initial data structures.  */
  ptu_lists = malloc_check (sizeof (struct ptu *) * packet_trace_file_count);
  current_host = host_list;

  /* Calls Log reader module.  */
  for (i = 0; i < packet_trace_file_count; i++)
    {
      read_log (current_host, &ptu_lists[i]);
      current_host = current_host->next;
    }

  /* Calls Message mapper module.  */
  ptp = map_messages (ptu_lists);

#ifdef DEBUG_MM
  fprintf (stderr, "=============== source list ================\n");
  print_debug_info2 (ptp);
#endif /* DEBUG_MM */
  /* Delete unused and make initial data structures.  */
  free (ptu_lists);
  es = NULL;			/* important for calculator */
/*     split (ptp, &es, host_list);*/
  /* Calls Event calculator module.  */
  calculate (ptp, &es, &link_list, &flow_list, variables_log_name);

#ifdef DEBUG_MM
  fprintf (stderr, "=============== result list ================\n");
  print_debug_info (es, host_list, flow_list, link_list);
#endif /* DEBUG_MM */
  exit (EXIT_SUCCESS);
}
