/*
 * File:   clp.c
 * Author: Andrew Salo
 * Date:   26.04.2004
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/types.h>

#include "analyzer.h"
#include "cmdlparse.h"
#include "error.h"
#include "logread.h"
#include "config.h"


/*
 * Print debugging information.
 */
void
lr_debug (struct ptu *begin)
{
  struct arphdr *arp_hdr;
  struct iphdr *ip_hdr;
  struct tcphdr *tcp_hdr;
  struct udphdr *udp_hdr;

  struct in_addr ip1, ip2;
  char strip1[INET_ADDRSTRLEN], strip2[INET_ADDRSTRLEN];

  char *data;
  int i;

  while (begin != NULL)
    {
      /* Print network layer protocol information.  */
      fprintf (stderr, "\nNet: ");

      switch (begin->net_type)
	{
	case PTU_ARP:
	  /* Network layer protocol is ARP.  */
	  arp_hdr = (struct arphdr *) begin->net_hdr;

	  get_ip_addrs ((void *) arp_hdr, PTU_ARP, &ip1, &ip2);

	  if (inet_ntop (AF_INET, &ip1, strip1, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  if (inet_ntop (AF_INET, &ip2, strip2, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  /* Print ARP header fields and IP addresses.  */
	  fprintf (stderr, "ARP [ar_hrd=%u, ar_pro=0x%x, ar_hln=%u, "
		   "ar_pln=%u, ar_op=%u, ar_sip=%s, ar_tip=%s]\n",
		   arp_hdr->ar_hrd, arp_hdr->ar_pro, arp_hdr->ar_hln,
		   arp_hdr->ar_pln, arp_hdr->ar_op, strip1, strip2);
	  break;

	case PTU_IPV4:
	  /* Network layer protocol is IPV4.  */
	  ip_hdr = (struct iphdr *) begin->net_hdr;

	  get_ip_addrs ((void *) ip_hdr, PTU_IPV4, &ip1, &ip2);

	  if (inet_ntop (AF_INET, &ip1, strip1, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  if (inet_ntop (AF_INET, &ip2, strip2, INET_ADDRSTRLEN) == NULL)
	    error (EXIT_FAILURE, "inet_ntop", strerror (errno));

	  /* Print IP header fields.  */
	  fprintf (stderr, "IPV4 [version=%u, ihl=%u, tos=%u, "
		   "tot_len=%u, id=%u, frag_off=0x%x, ttl=%u, "
		   "protocol=%u, check=0x%x, saddr=%s, daddr=%s]\n",
		   ip_hdr->version, ip_hdr->ihl, ip_hdr->tos,
		   ip_hdr->tot_len, ip_hdr->id, ip_hdr->frag_off,
		   ip_hdr->ttl, ip_hdr->protocol, ip_hdr->check,
		   strip1, strip2);
	  break;

	default:
	  fprintf (stderr, "unknown\n");
	}

      /* Print transport layer protocol information.  */
      fprintf (stderr, "Trans: ");

      switch (begin->trans_type)
	{
	case PTU_TCP:
	  /* Transport layer protocol is TCP.  */
	  tcp_hdr = (struct tcphdr *) begin->trans_hdr;

	  /* Print TCP header fields.  */
	  fprintf (stderr, "TCP [source=%u, dest=%u, seq=%u, "
		   "ack_seq=%u, doff=%u, urg=%u, ack=%u, "
		   "psh=%u, rst=%u, syn=%u, fin=%u, "
		   "window=%u, check=0x%x, urg_ptr=%u]\n",
		   tcp_hdr->source, tcp_hdr->dest, tcp_hdr->seq,
		   tcp_hdr->ack_seq, tcp_hdr->doff, tcp_hdr->urg,
		   tcp_hdr->ack, tcp_hdr->psh, tcp_hdr->rst, tcp_hdr->syn,
		   tcp_hdr->fin, tcp_hdr->window, tcp_hdr->check,
		   tcp_hdr->urg_ptr);
	  break;

	case PTU_UDP:
	  /* Transport layer protocol is UDP.  */
	  udp_hdr = (struct udphdr *) begin->trans_hdr;

	  /* Print TCP header fields.  */
	  fprintf (stderr, "UDP [source=%u, dest=%u, "
		   "len=%u, check=0x%x]\n",
		   udp_hdr->source, udp_hdr->dest, udp_hdr->len,
		   udp_hdr->check);
	  break;

	default:
	  fprintf (stderr, "unknown\n");
	}

      /* Print application layer protocol information.  */
      fprintf (stderr, "App: ");

      switch (begin->app_type)
	{
	case PTU_HTTP:
	  /* Application layer protocol is HTTP.  */
	  fprintf (stderr, "HTTP [len=%u]\n", begin->app_len);
	  
	  data = (char *) begin->app_data;

	  /* Print part of HTTP message.  */
	  for (i = 0; i < begin->app_len; i++)
	    fputc (data[i], stderr);
	  break;

	case PTU_DNS:
	  /* Application layer protocol is DNS.  */
	  fprintf (stderr, "DNS [len=%u]\n", begin->app_len);
	  break;

	default:
	  fprintf (stderr, "unknown [len=%u]\n", begin->app_len);
	}

      /* Move to next element.  */
      begin = begin->next;
    }
}


/*
 * Entry point of testing program.
 */
int
__wrap_main (int argc, char **argv)
{
  struct ptu *begin;
  struct host *tmp;
  int i;

  /* Parse command line arguments.  */
  parse_args (argc, argv);

  /* Read log files.  */
  for (i = 0, tmp = host_list; i < packet_trace_file_count; i++)
    {
      read_log (tmp, &begin);

#ifdef LR_DEBUG
      /* Print some debugging information.  */
      fprintf (stderr, "\nPACKET TRACE FILE: %s\n", tmp->ptfname);
      lr_debug (begin);
#endif /* LR_DEBUG */

      tmp = tmp->next;
    }

  exit (EXIT_SUCCESS);
}
